<?php

	// If WP_List_Table class doesn't exist, load it
	if(!class_exists('WP_List_Table')) {

		require_once(ABSPATH . 'wp-admin/includes/class-wp-list-table.php');
	}

	class WS_Form_WP_List_Table_Form extends WP_List_Table {

		// Construct
	    public function __construct() {

			parent::__construct(array(

				'singular'		=> __('Form', 'ws-form'), //Singular label
				'plural'		=> __('Forms', 'ws-form'), //plural label, also this well be one of the table css class
				'ajax'			=> false //We won't support Ajax for this table
			));
	    }

	    // Get columns
		public function get_columns() {

  		  	$columns = [

				'cb'			=> '<input type="checkbox" />',
				'label'			=> __('Name', 'ws-form'),
				'status'		=> __('Status', 'ws-form'),
				'count_submit'	=> __('Submissions', 'ws-form'),
				'shortcode'		=> __('Shortcode', 'ws-form')
			];

			return $columns;
		}

		// Get sortable columns
		public function get_sortable_columns() {

			$sortable_columns = array(

				'label'			=> array('label', true),
				'status'		=> array('status', true),
				'count_submit'	=> array('count_submit', true),
				'shortcode'		=> array('id', true)
			);

			return $sortable_columns;
		}

		// Column - Default
		public function column_default($item, $column_name) {

			switch ($column_name) {

				case 'name':
				case 'count_submit':

					return $item[$column_name];
					break;

				default:

					return print_r($item, true); //Show the whole array for troubleshooting purposes
			}
		}

		// Column - Checkbox
		function column_cb($item) {

			return sprintf('<input type="checkbox" name="bulk-ids[]" value="%u" />', $item['id']);
		}

		// Column - Label
		function column_label($item) {

			// URLs
			$id = absint($item['id']);
			$url_edit = WS_Form_Common::get_admin_url('ws-form-edit', $id);

			// Title
			if(WS_Form_Common::can_user('edit_form')) {

				$title = sprintf('<strong><a href="%s">%s</a></strong>', $url_edit, $item['label']);

			} else {

				$title = sprintf('<strong>%s</strong>', $item['label']);
			}

			// Actions
			$status = WS_Form_Common::get_query_var('ws-form-status');
			$actions = array();
			switch($status) {

				case 'trash' :

					if(WS_Form_Common::can_user('delete_form')) {

						$actions['restore'] = 	sprintf('<a href="#" data-action="wsf-restore" data-id="%u">%s</a>', $id, __('Restore', 'ws-form'));
						$actions['delete'] = 	sprintf('<a href="#" data-action="wsf-delete" data-id="%u">%s</a>', $id, __('Delete Permanently', 'ws-form'));
					}

					break;

				default :

					if(WS_Form_Common::can_user('edit_form')) {

						$actions['edit'] = 	sprintf('<a href="%s">%s</a>', $url_edit, __('Edit', 'ws-form'));
					}

					if(WS_Form_Common::can_user('create_form')) {

						$actions['copy'] = 	sprintf('<a href="#" data-action="wsf-clone" data-id="%u">%s</a>', $id, __('Clone', 'ws-form'));
					}

					if(WS_Form_Common::can_user('delete_form')) {

						$actions['trash'] = sprintf('<a href="#" data-action="wsf-delete" data-id="%u">%s</a>', $id, __('Trash', 'ws-form'));
					}

					$actions['preview'] = sprintf('<a href="%s" target="_blank">%s</a>', get_site_url(null, '/?wsf_preview_form_id=' . $id), __('Preview', 'ws-form'));
			}

			return $title . $this->row_actions($actions);
		}

		// Column - Status
		function column_status($item) {

			// Title
			$ws_form_form = New WS_Form_Form();
			$title = $ws_form_form->db_get_status_name($item['status']);

			return $title;
		}

		// Column - Submit Count
		function column_count_submit($item) {

			// URLs
			$id = absint($item['id']);
			$url_submissions = WS_Form_Common::get_admin_url('ws-form-submit&id=' . $id);

			// Get count (everything but trash)
			$ws_form_submit = new WS_Form_Submit();
			$count_submit = absint($ws_form_submit->db_get_count_by_status($id));

			if($count_submit > 0) {

				if(WS_Form_Common::can_user('read_submission')) {

					$title = sprintf('<a href="%s">%u</a>', $url_submissions, $count_submit);
	
				} else {

					$title = $count_submit;
				}
			} else {

				$title = '0';
			}

			// Actions
			$actions = array();
			if($count_submit > 0) {

				if(WS_Form_Common::can_user('read_submission')) {

					$actions['view'] = 		sprintf('<a href="%s">%s</a>', $url_submissions, __('View', 'ws-form'));
				}

				if(WS_Form_Common::can_user('export_submission')) {

					$actions['export'] = 		sprintf('<a href="%s">%s</a>', $url_submissions, __('Export', 'ws-form'));
				}
			}

			return $title . $this->row_actions($actions);
		}

		// Column - Shortcode
		function column_shortcode($item) {

			$id = absint($item['id']);

			// Title
			$title = sprintf('<code>%s</code>', WS_Form_Common::shortcode($id));

			return $title;
		}

		// Views
		function get_views(){

			// Get data from API
			$ws_form_form = New WS_Form_Form();

			$views = array();
			$current = ( !empty($_REQUEST['ws-form-status']) ? $_REQUEST['ws-form-status'] : 'all');

			// All link
			$count_all = $ws_form_form->db_get_count_by_status();
			if($count_all) {
				$class = ($current === 'all' ? ' class="current"' :'');
				$all_url = remove_query_arg('ws-form-status');
				$views['all'] = "<a href=\"{$all_url}\" {$class} >" . __('All', 'ws-form') . "</a> <span class=\"count\">($count_all)</span>";
			}

			// Draft link
			$count_draft = $ws_form_form->db_get_count_by_status('draft');
			if($count_draft) {
				$draft_url = add_query_arg('ws-form-status','draft');
				$class = ($current === 'draft' ? ' class="current"' :'');
				$views['draft'] = "<a href=\"{$draft_url}\" {$class} >" . __('Draft', 'ws-form') . "</a> <span class=\"count\">($count_draft)</span>";
			}

			// Published link
			$count_publish = $ws_form_form->db_get_count_by_status('publish');
			if($count_publish) {
				$publish_url = add_query_arg('ws-form-status','publish');
				$class = ($current === 'publish' ? ' class="current"' :'');
				$views['publish'] = "<a href=\"{$publish_url}\" {$class} >" . __('Published', 'ws-form') . "</a> <span class=\"count\">($count_publish)</span>";
			}

			// Trashed link
			$count_trash = $ws_form_form->db_get_count_by_status('trash');
			if($count_trash) {
				$trash_url = add_query_arg('ws-form-status','trash');
				$class = ($current === 'trash' ? ' class="current"' :'');
				$views['trash'] = "<a href=\"{$trash_url}\" {$class} >" . __('Trash', 'ws-form') . "</a> <span class=\"count\">($count_trash)</span>";
			}

			return $views;
		}

		// Get data
		function get_data($per_page = 20, $page_number = 1) {

			// Build JOIN
			$join = '';

			// Build WHERE
			$status = WS_Form_Common::get_query_var('ws-form-status');
			if($status == '') { $status == 'all'; }
			if(!WS_Form_Common::check_form_status($status, false)) { $status = 'all'; }
			if($status != 'all') {
	
				// Filter by status
				$where = 'status = "' . $status . '"';

			} else {

				// Show everything but trash (All)
				$where = " NOT(status = 'trash')";
			}

			// Build ORDER BY
			$order_by = '';
			if (!empty($_REQUEST['orderby'])) {

				$order_by = esc_sql($_REQUEST['orderby']);
				$order_by .= !empty($_REQUEST['order']) ? ' ' . esc_sql($_REQUEST['order']) : ' ASC';
			} else {

				$order_by = 'id DESC';
			}

			// Build LIMIT
			$limit = $per_page;

			// Build OFFSET
			$offset = ($page_number - 1) * $per_page;

			// Get data from API
			$ws_form_form = New WS_Form_Form();
			$result = $ws_form_form->db_read_all($join, $where, $order_by, $limit, $offset);

			return $result;
		}

		// Prepare items
		public function prepare_items() {

			$columns = $this->get_columns();
// 			$hidden = $this->get_hidden_columns();
			$sortable = $this->get_sortable_columns();

			$this->_column_headers = array($columns, array(), $sortable);

			$per_page     = $this->get_items_per_page('ws_form_items_per_page_form', 20);
			$current_page = $this->get_pagenum();
			$total_items  = self::record_count();

			$this->set_pagination_args(array(

				'total_items' => $total_items, //WE have to calculate the total number of items
				'per_page'    => $per_page //WE have to determine how many items to show on a page
			));

			$this->items = self::get_data($per_page, $current_page);
		}

		// Bulk actions - Prepare
		public function get_bulk_actions() {

			$actions = array();
			$status = WS_Form_Common::get_query_var('ws-form-status');

			switch($status) {

				case 'trash' :

					if(WS_Form_Common::can_user('delete_form')) {

						$actions['wsf-bulk-restore'] = __('Restore', 'ws-form');
						$actions['wsf-bulk-delete'] = __('Delete Permanently', 'ws-form');
					}
					break;

				default:

					if(WS_Form_Common::can_user('delete_form')) {

						$actions['wsf-bulk-delete'] = __('Move to Trash', 'ws-form');
					}
			}

			return $actions;
		}

		// Extra table nav
		function extra_tablenav( $which ) {

			$status = WS_Form_Common::get_query_var('ws-form-status');

			switch($status) {

				case 'trash' :

					if(WS_Form_Common::can_user('delete_form')) {
?>
		<div class="alignleft actions">
<?php 
			submit_button(__('Empty Trash', 'ws-form'), 'apply', 'delete_all', false );
?>
		</div>
<?php
					}

					break;
			}
		}

		// Get record count
		public function record_count() {

			$ws_form_form = New WS_Form_Form();

			$current = ( !empty($_REQUEST['ws-form-status']) ? $_REQUEST['ws-form-status'] : 'all');
			if($current === 'all') { $current = ''; }

			return $ws_form_form->db_get_count_by_status($current);
		}

		// No records
		public function no_items() {

			_e('No forms avaliable.', 'ws-form');
		}
	}
