<?php

	class WS_Form_API_Helper extends WS_Form_API {

		public function __construct() {

			// Call parent on WS_Form_API
			parent::__construct();
		}

		// API - Detect framework
		public function api_framework_detect($parameters) {

			// Get file path if provided
			$path = WS_Form_Common::get_query_var('path', '', $parameters);

			// Get framework auto detect configuration
			$frameworks = WS_Form_Config::get_frameworks(false);
			if(!isset($frameworks['auto_detect'])) { self::api_throw_error(__('Framework detect configuration not found', 'ws-form')); }

			$auto_detect = $frameworks['auto_detect'];

			// Get framework type lookups
			$types = (isset($auto_detect['types'])) ? $auto_detect['types'] : false;
			if($types === false) { self::api_throw_error(__('Framework detect types configuration not found', 'ws-form')); }

			// Get framework filename exclusions
			$exclude_filenames = (isset($auto_detect['exclude_filenames'])) ? $auto_detect['exclude_filenames'] : false;

			// Get website URL
			$url = site_url($path);
			if(!$url) { return false; }

			// Make HTTP request to get URL
			$response = wp_remote_get($url);
			if(!is_array($response)) { self::api_throw_error(__('Framework detect HTTP request error', 'ws-form')); }
			if(!isset($response['body'])) { self::api_throw_error(__('Framework detect HTTP request error', 'ws-form')); }

			// Read body resposne
			$http_body = $response['body']; // use the content
			if($http_body == '') { self::api_throw_error(__('Framework detect HTTP body blank', 'ws-form')); }
			if((strpos($http_body, 'css') === false) && (strpos($http_body, 'CSS') === false)) { self::api_throw_error(__('Framework detect HTTP body does not contain CSS references', 'ws-form')); }

			// Start DOM document
			$dom_doc = new DOMDocument();

			// Load HTML into DOM document (diseregard parse errors)
			libxml_use_internal_errors(true);
			if(!$dom_doc->loadHTML($http_body)) { self::api_throw_error(__('Framework detect HTML load error', 'ws-form')); }
			libxml_use_internal_errors(false);

			// Look for link tags
			$domcss = $dom_doc->getElementsByTagName('link');
			foreach($domcss as $links) {

				// Look for rel attributes
				if(strtolower($links->getAttribute('rel')) == "stylesheet" ) {

					// Get href attribute
					$css_url = $links->getAttribute('href');

					// Do we recognize the file name?
					if($exclude_filenames !== false) {

						foreach($exclude_filenames as $exclude_filename) {

							if(strpos($css_url, $exclude_filename) !== false) { continue; }
						}
					}

					// Request CSS document
					$css_response = wp_remote_get($css_url);

					// Load response body into string
					$css_body = $css_response['body'];

					// Run through each framework type
					foreach($types as $type => $type_strings) {

						$lookup_strings_found = true;

						// Run through each string to find in the framework
						foreach($type_strings as $type_string) {

							// Look for element in CSS body (Case sensitive)
							if(strpos($css_body, $type_string) === false) {

								$lookup_strings_found = false;
								break;
							}
						}

						// If all strings are found, return that framework
						if($lookup_strings_found) {

							// Return framework data
							$return_array = array();
							$return_array['type'] = $type;
							$return_array['framework'] = $frameworks['types'][$type];
							self::api_json_response($return_array, 0, false);
						}
					}
				}
			}

			// Unable to find a framework
			self::api_throw_error(__('Framework detection failed', 'ws-form'));
		}

		// API - Push setup
		public function api_setup_push($parameters) {

			// User capability check
			if(!WS_Form_Common::can_user('manage_options_wsform')) { return false; }

			// Get framework
			$framework = WS_Form_Common::get_query_var('framework', '', $parameters);
			if($framework == '') { self::api_throw_error(__('Framework not specified', 'ws-form')); }

			// Check framework
			$frameworks = WS_Form_Config::get_frameworks(false);
			if(!isset($frameworks['types'][$framework])) { self::api_throw_error(__('Invalid framework specified', 'ws-form')); }

			// Get mode
			$mode = WS_Form_Common::get_query_var('mode', '', $parameters);
			if($mode == '') { $mode = WS_FORM_DEFAULT_MODE; }

			// Check mode
			$modes = explode(',', WS_FORM_MODES);
			if(!in_array($mode, $modes)) { self::api_throw_error(__('Invalid mode specified', 'ws-form')); }

			// Set framework
			WS_Form_Common::option_set('framework', $framework);

			// Set mode
			WS_Form_Common::option_set('mode', $mode);

			// Configure settings according to mode selected
			$options = WS_Form_Config::get_options();
			foreach($options as $tab => $data) {

				if(isset($data['fields'])) {

					$fields = $data['fields'];
				}

				if(isset($data['groups'])) {

					$groups = $data['groups'];

					foreach($groups as $group) {

						$fields = $group['fields'];

						self::api_set_push_options($mode, $fields);
					}
				}
			}

			// Set setup (true = complete)
			WS_Form_Common::option_set('setup', true);

			// Success
			self::api_json_response([], 0, false);
		}

		// API - Push setup - Set options
		public function api_set_push_options($mode, $fields) {

			foreach($fields as $key => $attributes) {

				if(
					isset($attributes['type']) && 
					($attributes['type'] != 'static') && 
					isset($attributes['mode']) &&
					isset($attributes['mode'][$mode])
				) {

					$value = $attributes['mode'][$mode];

					WS_Form_Common::option_set($key, $value);
				}
			}
		}

		// API - Support contact submit
		public function api_support_contact_submit() {

			// Read support inquiry fields
			$data = array(

				'contact_first_name'	=> WS_Form_Common::get_query_var('contact_first_name'),
				'contact_last_name'		=> WS_Form_Common::get_query_var('contact_last_name'),
				'contact_email'			=> WS_Form_Common::get_query_var('contact_email'),
				'contact_inquiry'		=> WS_Form_Common::get_query_var('contact_inquiry')
			);

			// Push form
			$contact_push_form = WS_Form_Common::get_query_var('contact_push_form');
			$form_id = absint(WS_Form_Common::get_query_var('form_id'));
			if($contact_push_form && ($form_id > 0)) {

				// Create form file attachment
				$ws_form_form = new WS_Form_Form();
				$ws_form_form->id = $form_id;

				// Get form
				$form = $ws_form_form->db_read(true, true);

				// Convert to object
				$form = json_decode(json_encode($form));

				// Clean form
				unset($form->checksum);
				unset($form->published_checksum);

				// Stamp form data
				$form->identifier = WS_FORM_IDENTIFIER;
				$form->version = WS_FORM_VERSION;
				$form->time = time();

				// Add checksum
				$form->checksum = md5(json_encode($form));

				$form_json = json_encode($form);

				// Add to data
				$data['contact_form'] = $form_json;
			}

			// Push system
			$contact_push_system = WS_Form_Common::get_query_var('contact_push_system');
			if($contact_push_system) {

				// Add to data
				$data['contact_system'] = json_encode(WS_Form_Config::get_system());
			}

			// Filters
			$timeout = apply_filters('wsf_api_call_timeout', WS_FORM_API_CALL_TIMEOUT);
			$sslverify = apply_filters('wsf_api_call_verify_ssl',WS_FORM_API_CALL_VERIFY_SSL);

			// Build args
			$args = array(

				'method'		=>	'POST',
				'body'			=>	http_build_query($data),
				'user-agent'	=>	'WSForm/' . WS_FORM_VERSION . ' (wsform.com)',
				'timeout'		=>	$timeout,
				'sslverify'		=>	$sslverify
			);

			// URL
			$url = 'https://wsform.com/plugin-support/contact.php';

			// Call using Wordpress wp_remote_get
			$response = wp_remote_get($url, $args);

			// Check for error
			if($api_response_error = is_wp_error($response)) {

				// Handle error
				$api_response_error_message = $response->get_error_message();;
				$api_response_headers = array();
				$api_response_body = '';
				$api_response_http_code = 0;

			} else {

				// Handle response
				$api_response_error_message = '';
				$api_response_headers = wp_remote_retrieve_headers($response);
				$api_response_body = wp_remote_retrieve_body($response);
				$api_response_http_code = wp_remote_retrieve_response_code($response);
			}

			// Return response
			echo json_encode(array('error' => $api_response_error, 'error_message' => $api_response_error_message, 'response' => $api_response_body, 'http_code' => $api_response_http_code));

			exit;
		}

		// API - WS Form Admin CSS
		public function api_ws_form_css_admin() {

			// Output HTTP header
			header("Content-type: text/css; charset: UTF-8");

			// Output CSS
			$ws_form_css = new WS_Form_CSS();
			$css = $ws_form_css->get_admin();

			echo $css;

			exit;
		}

		// API - WS Form Public CSS
		public function api_ws_form_css() {

			header("Content-type: text/css; charset: UTF-8");

			// Output CSS
			$ws_form_css = new WS_Form_CSS();
			$css = $ws_form_css->get_public();

			echo $css;

			exit;
		}

		// API - WS Form Skin CSS
		public function api_ws_form_css_skin() {

			// Output HTTP header
			header("Content-type: text/css; charset: UTF-8");

			// Output CSS
			$ws_form_css = new WS_Form_CSS();
			$ws_form_css->render_skin();

			exit;
		}

		// API - Detect framework
		public function api_form_preview($parameters) {

			// Enable wizard mode
			global $wizard;
			$wizard = true;

			// Get wizard ID
			$wizard_id = intval(WS_Form_Common::get_query_var('id', '', $parameters));
			if($wizard_id == 0) { exit; }

			// Read wizard data
			$ws_form_wizard = new WS_Form_Wizard();
			$ws_form_wizard->id = $wizard_id;
			$ws_form_wizard->db_read();

			// Public data
			$ws_form_public = new WS_Form_Public(WS_FORM_NAME, WS_FORM_VERSION);
			$ws_form_settings = json_encode($ws_form_public->localization_object(false));
			$ws_form_html = $ws_form_public->form_html(1, 1, $ws_form_wizard->form);

			// Set content type
			header("Content-type: text/html; charset: UTF-8");
?>
<html>

<head>

<title>Form Preview</title>

<style>

body { padding: 40px; }

</style>

<link rel='stylesheet' href='<?php echo WS_Form_Common::get_api_base_path() . 'helper/ws_form_css'; ?>' type='text/css' media='all' />
<link rel='stylesheet' href='<?php echo WS_Form_Common::get_api_base_path() . 'helper/ws_form_css_skin?wizard=1'; ?>' type='text/css' media='all' />

<script type='text/javascript' src='<?php echo includes_url('js/jquery/jquery.js'); ?>'></script>

<script>

	// Localization
	var ws_form_settings = <?php echo $ws_form_settings; ?>;

</script>

<script type='text/javascript' src='<?php plugin_dir_url(__DIR__) . 'shared/js/ws-form.js'; ?>'></script>
<script type='text/javascript' src='<?php plugin_dir_url(__DIR__) . 'public/js/ws-form-public.js'; ?>'></script>

</head>

<body>
<?php

	// Render form HTML
	echo $ws_form_html;
?>
</body>

</html>
<?php
			exit;
		}

		// API - File download
		public function api_file_download($parameters) {

			// User capability check
			if(!WS_Form_Common::can_user('read_submission')) { return false; }

			// Get submit hash
			$hash = WS_Form_Common::get_query_var('hash', '', $parameters);
			if($hash == '' || (strlen($hash) != 32)) { self::api_throw_error(__('Hash not specified', 'ws-form')); }

			// Get field ID
			$field_id = intval(WS_Form_Common::get_query_var('field_id', '', $parameters));
			if($field_id == 0) { self::api_throw_error(__('Field ID not specified', 'ws-form')); }

			// Get file index
			$file_index = intval(WS_Form_Common::get_query_var('file_index', '', $parameters));
			if($file_index < 0) { self::api_throw_error(__('File index invalid', 'ws-form')); }

			// Get submit record
			$ws_form_submit = new WS_Form_Submit();
			$ws_form_submit->hash = $hash;
			$submit = $ws_form_submit->db_read_by_hash(true, false, false);

			// Get field
			if(!isset($submit['meta'][WS_FORM_FIELD_PREFIX . $field_id])) { self::api_throw_error(__('Field ID not found', 'ws-form')); }
			$field = $submit['meta'][WS_FORM_FIELD_PREFIX . $field_id];

			// Get files
			if(!isset($field['value'])) { self::api_throw_error(__('Field data not found', 'ws-form')); }
			$files = $field['value'];

			// Get file
			if(!isset($files[$file_index])) { self::api_throw_error(__('Field data not found', 'ws-form')); }
			$file = $files[$file_index];

			// Get file name
			if(!isset($file['name'])) { self::api_throw_error(__('File name not found', 'ws-form')); }
			$file_name = $file['name'];

			// Get file type
			if(!isset($file['type'])) { self::api_throw_error(__('File type not found', 'ws-form')); }
			$file_type = $file['type'];

			// Get file path
			if(!isset($file['path'])) { self::api_throw_error(__('File path not found', 'ws-form')); }
			$file_path = $file['path'];

			// Set HTTP headers
			header('Content-Type: ' . $file_type);

			// Make browser download file instead of viewing it
			$download = (WS_Form_Common::get_query_var('download', '', $parameters) !== '');
			if($download) {

				header("Content-Transfer-Encoding: Binary"); 
				header("Content-disposition: attachment; filename=\"" . $file_name . "\""); 
			}

			// Get base upload_dir
			$upload_dir = wp_upload_dir()['basedir'];

			// Build file path
			$file_path_full = $upload_dir . '/' . $file_path;

			// Push file to browser
			readfile($file_path_full);

			exit;
		}

		// Hidden columns changed via AJAX request
		public function api_user_meta_hidden_columns($parameters) {

			// User capability check
			if(!WS_Form_Common::can_user('read_submission')) { return false; }

			$form_id = intval(WS_Form_Common::get_query_var('form_id', '', $parameters));
			if($form_id == 0) { exit; }

			// Get hidden columns
			$form_hidden_columns_string = WS_Form_Common::get_query_var('hidden', '', $parameters);
			$form_hidden_columns = explode(',', $form_hidden_columns_string);

			// Get current user
			$user = wp_get_current_user();

			// Write hidden columns back to user meta for current form
			update_user_option($user->ID, 'managews-form_page_ws-form-submitcolumnshidden-' . $form_id, $form_hidden_columns, true);

			self::api_json_response();
		}

		// Get variables for TinyMCE editor ListBox
		public function api_variables($parameters) {

			// Get form ID
			$form_id = intval(WS_Form_Common::get_query_var('form_id', 0, $parameters));
			if($form_id == 0) { exit; }

			$variables = WS_Form_Config::get_parse_variable_help($form_id, false);

			echo json_encode($variables);
			exit;
		}
	}