<?php

	/**
	 * Common functions used by this plugin
	 */
	class WS_Form_Common {

		// Encryption key, we recommend changing this
		const WS_FORM_COOKIE_PREFIX = 'ws_form_';

		// Admin messages
		public static $admin_messages = array();

		// IP lookup response
		public static $ip_lookup_response = false;

		// Admin messages - Push
		public static function admin_message_push($message, $type = 'notice-success', $dismissible = true, $nag_notice = false) {

			self::$admin_messages[] = array(

				'message'		=>	$message,
				'type'			=>	$type,
				'dismissible'	=>	$dismissible,
				'nag_notice'	=>	$nag_notice
			);
		}

		// Admin messages - Render
		public static function admin_messages_render() {

			// Server side notices
			foreach(self::$admin_messages as $admin_message) {

				$message = $admin_message['message'];
				$type = isset($admin_message['type']) ? $admin_message['type'] : 'notice-success';
				$dismissible = isset($admin_message['dismissible']) ? $admin_message['dismissible'] : true;
				$nag_notice = isset($admin_message['nag_notice']) ? $admin_message['nag_notice'] : false;

				self::admin_message_render($message, $type, $dismissible, $nag_notice);
			}
		}

		// Admin messages - Render single
		public static function admin_message_render($message, $type = 'notice-success', $dismissible = true, $nag_notice = false) {

			if(!(defined('DISABLE_NAG_NOTICES') && DISABLE_NAG_NOTICES && $nag_notice)) {

				$notice = '<div class="notice ' . $type . (($dismissible) ? ' is-dismissible' : '') . '"><p>' . str_replace("\n", "<br />\n", $message) . '</p></div>';

				echo $notice;
			}
		}

		// Admin messages - Get count
		public static function get_admin_message_count() {

			return count(self::$admin_messages);
		}

		// Get plugin option key value
		public static function option_get($key, $default = false, $default_set = false) {

			$options = get_option(WS_FORM_IDENTIFIER, false);
			$options_array = (is_array($options)) ? $options : array();

			// Wizard overrides
			global $wizard;
			if($wizard && ($key == 'framework')) { return WS_FORM_DEFAULT_FRAMEWORK; }

			// If key exists, return the value
			if(isset($options_array[$key])) {

				return $options_array[$key];

			} else {

				// Set value
				if($default_set) { self::option_set($key, $default); }

				// Return default
				return $default;
			}
		}

		// Set plugin option key value
		public static function option_set($key, $value, $update = true) {

			$options = get_option(WS_FORM_IDENTIFIER, false);
			$options_array = (is_array($options)) ? $options : array();

			if((isset($options_array[$key]) && $update) || (!isset($options_array[$key]))) {

				// Set key to value in options array
				$options_array[$key] = $value;
			}

			// Update WordPress option
			update_option(WS_FORM_IDENTIFIER, $options_array);
		}

		// Remove plugin option key value
		public static function option_remove($key) {

			$options = get_option(WS_FORM_IDENTIFIER, $default);
			$options_array = (is_array($options)) ? $options : array();

			// If key exists, return the value
			if(isset($options_array[$key])) {

				// Remove key
				unset($options_array[$key]);

				// Update WordPress option
				update_option(WS_FORM_IDENTIFIER, $options_array);

				// Key found and removed
				return true;
			}

			// Did not find key
			return false;
		}

		// Get admin URL
		public static function get_admin_url($page_slug = '', $item_id = false, $path_extra = '') {

			$page_path = 'admin.php';
			if($page_slug != '') { $page_path .= '?page=' . $page_slug; }
			if($item_id !== false) { $item_id = intval($item_id); $page_path .= '&id=' . $item_id; }
			if($path_extra) { $page_path .= '&' . $path_extra; }

			return admin_url($page_path);
		}

		// Get plugin website link
		public static function get_plugin_website_url($path = '', $medium = false) {

			return sprintf('https://wsform.com%s?utm_source=ws_form%s', $path, (($medium !== false) ? '&utm_medium=' . $medium : ''));
		}

		// Get query var
		public static function get_query_var($var, $default = '', $parameters = false, $esc_sql = false, $strip_slashes = true) {

			// REST parameters
			if($parameters !== false) {

				if(isset($parameters[$var])) {

					$return_value = $esc_sql ? esc_sql($parameters[$var]) : $parameters[$var];
					return $strip_slashes ? stripslashes_deep($return_value) : $return_value;
				}
			}

			// Get from standard _GET _POST arrays
			$request_method = WS_Form_Common::get_request_method();
			if(!$request_method) { return $default; }

			// Check if X-HTTP-Method-Override is being used
			$ajax_http_method_override = self::option_get('ajax_http_method_override', true);
			if(!$ajax_http_method_override && ($request_method !== 'POST')) {

				parse_str(file_get_contents('php://input'), $_POST);
				$strip_slashes = false;
			}

			// DATA param (This overcomes standard 1000 POST parameter limitation in PHP)
			if(isset($_POST)) {

				if(isset($_POST['data'])) {

					$data = $strip_slashes ? stripslashes_deep($_POST['data']) : $_POST['data'];

					$data_array = is_string($data) ? json_decode($data, true) : array();

					if(isset($data_array[$var])) { return $data_array[$var]; }
				}
			}

			// Regular GET, POST, PUT handling
			switch($request_method) {

				case 'GET' :

					$return_value = isset($_GET[$var]) ? ($esc_sql ? esc_sql($_GET[$var]) : $_GET[$var]) : $default;
					return $strip_slashes ? stripslashes_deep($return_value) : $return_value;

				case 'POST' :
				case 'PUT' :

					$return_value = isset($_POST[$var]) ? ($esc_sql ? esc_sql($_POST[$var]) : $_POST[$var]) : $default;
					return $strip_slashes ? stripslashes_deep($return_value) : $return_value;

				default :

					return $default;
			}
		}

		// Get IP lookup
		// Data requested from geoplugin.net that uses GeoLite data created by MaxMind, available from http://www.maxmind.com
		public static function get_ip_lookup($json_var_array) {

			if(!is_array($json_var_array)) { $json_var_array = array($json_var_array); }

			if(self::$ip_lookup_response === false) {

				// Get remote IP address
				$remote_addr = self::get_http_env(array('HTTP_CLIENT_IP', 'HTTP_X_FORWARDED_FOR', 'REMOTE_ADDR'));
				if($remote_addr == '') { return ''; }

				// Do IP Lookup
				self::$ip_lookup_response = file_get_contents(sprintf('http://www.geoplugin.net/php.gp?ip=%s', $remote_addr));
			}

			if(@unserialize(self::$ip_lookup_response) !== false) {

				$ip_lookup_array = unserialize(self::$ip_lookup_response);

				$return_array = array();

				foreach($json_var_array as $json_var) {

					if(isset($ip_lookup_array[$json_var])) {

						$return_array[] = $ip_lookup_array[$json_var];
					}
				}

				return implode(',', $return_array);
			}

			return '';
		}

		// Get MySQL date
		public static function get_mysql_date($date = false) {

			$time = ($date === false) ? time() : strtotime($date);
			if($time === false) { return false; }
			return date('Y-m-d H:i:s', $time);
		}

		// Get request method
		public static function get_request_method($valid_request_methods = false) {

			// Check for valid request methods
			if(!$valid_request_methods) { $valid_request_methods = ['GET', 'POST', 'PUT', 'DELETE']; }

			// Check to ensure we can determine request method
			if(!isset($_SERVER) || !isset($_SERVER["REQUEST_METHOD"])) { return false; }

			// Read request method
			$request_method = strtoupper($_SERVER["REQUEST_METHOD"]);

			// Ensure it is valid
			if(!in_array($request_method, $valid_request_methods)) { return false; }

			return ($request_method != '') ? $request_method : false;
		}

		// Get current user ID
		public static function get_user_id($exit_on_zero = true) {

			$user_id = get_current_user_id();
			if(($user_id == 0) && $exit_on_zero) { exit; }
			return($user_id);
		}

		// Get current URL
		public static function get_current_url() {

			if(!isset($_SERVER)) { return false; }

			return sprintf('%s://%s%s', (self::get_http_env('HTTPS') === 'on' ? "https" : "http"), self::get_http_env('HTTP_HOST'), self::get_http_env('REQUEST_URI'));
		}

		// Add query string to URL
		public static function add_query_string($url, $query_string) {

			$url_parsed = parse_url($url);
			if(!isset($url_parsed['path'])) { $url .= '/'; }
			$separator = isset($url_parsed['query']) ? '&' : '?';
			$url .= $separator . $query_string;

			return $url;
		}

		// Echo comment - CSS
		public static function comment_css($comment) {

			// Should CSS be commented?
			$comments_css = WS_Form_Common::option_get('comments_css', true);

			if($comments_css) {

				// Echo comment
				$return_css = sprintf("/* %s */\n", $comment);

			} else {

				$return_css = '';
			}

			return $return_css;
		}

		// SVG Render
		public static function render_icon_16_svg($id) {

			$return_html = WS_Form_Config::get_icon_16_svg($id);

			if($return_html !== false) { echo $return_html; }

			return $return_html;
		}

		// Check form status
		public static function check_form_status($status, $throw_error = true) {

			// Check status is valid
			$valid_statuses = explode(',', WS_FORM_STATUS_FORM);
			if(!in_array($status, $valid_statuses)) {

				if($throw_error) {

					self::throw_error(__('Attempt to set invalid status on form', 'ws-form'));

				} else {

					return false;
				}
			}

			return true;
		}

		// Check submit status
		public static function check_submit_status($status, $throw_error = true) {

			// Check status is valid
			$valid_statuses = explode(',', WS_FORM_STATUS_SUBMIT);
			if(!in_array($status, $valid_statuses)) {

				if($throw_error) {

					self::throw_error(__('Attempt to set invalid status on submit', 'ws-form'));

				} else {

					return false;
				}
			}

			return true;
		}

		// Set encrypted cookie
		public static function set_encrypted_cookie($cookieName, $value, $timeout = false, $delete = false) {

			// Encrypt value
			$valueEncrypted = self::encrypt($value);

			// Set timeout
			if(!$timeout) { $timeout = self::option_get('cookie_timeout', 3600); }

			// Delete?
			if($delete) {
				$cookieExpiry = time() - 3600;
			} else {
				$cookieExpiry = time() + $timeout;
			}

			// Set cookie
			setcookie(self::option_get('cookie_prefix') . '_' . $cookieName, $valueEncrypted, $cookieExpiry, '/');
		} 

		// Get encrypted cookie
		public static function get_ecrypted_cookie($cookieName, $default = false) {

			if(isset($_COOKIE[self::option_get('cookie_prefix') . $cookieName])) {

				$valueEncrypted = $_COOKIE[self::option_get('cookie_prefix') . '_' . $cookieName];
				if(!empty($valueEncrypted)) {

					$value = trim(self::decrypt($valueEncrypted));
					return $value;
				}
			}

			return $default;
		}

	    // Encrypt a string
	    private function encrypt($input) {

	    	if($input == '') { return ''; }

	        $data = mcrypt_encrypt(MCRYPT_RIJNDAEL_128, self::option_get('encrypt_key'), $input, MCRYPT_MODE_ECB, 'keee');

	        $returnValue = base64_encode($data);

	        return $returnValue;
	    }

	    // Decrypt a string
	    private function decrypt($input) {

	    	if($input == '') { return ''; }

	        $data = base64_decode($input);

	        $returnValue = mcrypt_decrypt(MCRYPT_RIJNDAEL_128, self::option_get('encrypt_key'), $data, MCRYPT_MODE_ECB, 'keee');

	        return $returnValue;
	    }

		// Get API base path
		public static function get_api_base_path() {

			// Check permalinks
			$api_base_path = (get_option('permalink_structure') == '/%postname%/') ? get_site_url() . '/wp-json/' : get_site_url() . '/?rest_route=/';
			$api_base_path .= WS_FORM_RESTFUL_NAMESPACE . '/';

			return $api_base_path;
		}

		// Is debug enabled?
		public static function debug_enabled() {

			switch(WS_Form_Common::option_get('helper_debug', 'off')) {

				case 'administrator' : 	

					if(function_exists('wp_get_current_user')) {

						$user = wp_get_current_user();
						return in_array('administrator', $user->roles);

					} else {

						return false;
					}
	
				case 'on' :

					return true;
			}

			return false;
		}

		// Add populate values to fields
		public static function get_fields_from_form($form_array) {

			$fields = self::get_group_fields($form_array->groups);

			return $fields;
		}

		// Run through each group
		public static function get_group_fields($groups) {

			$fields = array();

			foreach($groups as $key => $group) {

				if(isset($groups[$key]->sections)) {

					$section_fields = self::get_section_fields($group->sections);

					$fields = $fields + $section_fields;
				}
			}

			return $fields;
		}

		// Run through each section
		public static function get_section_fields($sections) {

			$fields = array();

			foreach($sections as $key => $section) {

				if(isset($sections[$key]->children)) {

					$sections[$key]->children = self::get_section_fields($section->children, $fields);
				}

				if(isset($sections[$key]->fields)) {

					$section_fields = array();

					foreach($section->fields as $field) {

						$section_fields[$field->id] = $field;
					}

					$fields = $fields + $section_fields;
				}
			}

			return $fields;
		}

		// Mask parse
		public static function mask_parse($mask, $values) {

			if($mask == '') { return ''; }

			foreach($values as $key => $value) {

				$mask = str_replace('#' . $key, $value, $mask);
			}

			return $mask;
		}

		// Create shortcode
		public static function shortcode($id = false) {

			if($id == false) { return ''; }

			$shortcode = sprintf('[%s id="%u"]', WS_FORM_SHORTCODE, $id);

			return $shortcode;
		}

		// Check file upload capabilities
		public static function uploads_check() {

			// Create file warnings
			$files_warning = [];

			// Read ini settings
			if(!ini_get('upload_max_filesize')) { return(['max_upload_size' => 0, 'errors' => [__('Unable to read PHP upload_max_filesize setting', 'ws-form')]]); }
			$upload_max_filesize = self::ini_shorthand_notation_to_bytes(ini_get('upload_max_filesize'));

			if(!ini_get('post_max_size')) { return(['max_upload_size' => 0, 'errors' => [__('Unable to read PHP post_max_size setting', 'ws-form')]]); }
			$post_max_size = self::ini_shorthand_notation_to_bytes(ini_get('post_max_size'));

			if(!ini_get('memory_limit')) { return(['max_upload_size' => 0, 'errors' => [__('Unable to read PHP memory_limit setting', 'ws-form')]]); }
			$memory_limit = self::ini_shorthand_notation_to_bytes(ini_get('memory_limit'));

			// This limit was added in PHP 5.3.1
			if(ini_get('max_file_uploads')) {
				$max_file_uploads = absint(ini_get('max_file_uploads'));
			} else {
				$max_file_uploads = 20;
			}

			// Calculate recommended maximum upload size
			$max_upload_size = $upload_max_filesize;
			if($max_upload_size > $post_max_size) { $max_upload_size = $post_max_size; }
			if($max_upload_size > $memory_limit) { $max_upload_size = $memory_limit; }

			// Check ini settings
			if($post_max_size < $upload_max_filesize) { $files_warning[] = sprintf(__('Your PHP post_max_size setting (%s) is less than your max_upload_size setting (%s).', 'ws-form'), ini_get('memory_limit'), ini_get('post_max_size')); }
			if($memory_limit < $upload_max_filesize) { $files_warning[] = sprintf(__('Your PHP memory_limit setting (%s) is less than your max_upload_size setting (%s).', 'ws-form'), ini_get('memory_limit'), ini_get('post_max_size')); }
			if($memory_limit < $post_max_size) { $files_warning[] = sprintf(__('Your PHP memory_limit setting (%s) is less than your post_max_size setting (%s).', 'ws-form'), ini_get('memory_limit'), ini_get('post_max_size')); }

			// Check file permissions
			$upload_dir_create = self::upload_dir_create();
			if($upload_dir_create['error']) { $files_warning[] = $upload_dir_create['message']; }

			// Return result
			return ['max_upload_size' => $max_upload_size, 'max_uploads' => $max_file_uploads, 'errors' => $files_warning];
		}

		// Make sure upload folder can be created. Create it if it doesn't exist.
		public static function upload_dir_create($dir = '') {

			// Get base upload_dir
			$upload_dir = wp_upload_dir()['basedir'];

			// Check upload directory can be written to
			if(!is_writeable($upload_dir)) { return ['error' => true, 'message' => __('Your WordPress uploads directory cannot be written to.', 'ws-form')]; }

			// Get 
			$upload_dir_path = WS_FORM_UPLOAD_DIR . (($dir != '') ? '/' . $dir : '');
			$upload_dir_ws_form = $upload_dir . '/' . $upload_dir_path;

			// Check to see if WS Forms upload folder exists
			if(!file_exists($upload_dir_ws_form)) {

				if(!wp_mkdir_p($upload_dir_ws_form)) { return ['error' => true, 'message' => sprintf(__('Unable to create upload folder for WS Form uploaded files (wp-content/uploads/%s).', 'ws-form'), $upload_dir_path)]; }
			}

			return ['error' => false, 'dir' => $upload_dir_ws_form, 'path' => $upload_dir_path];
		}

		// Convert PHP ini shorthand notation to bytes
		public static function ini_shorthand_notation_to_bytes($ini_value) {

			$ini_value  = trim($ini_value);

			$last = strtolower($ini_value[strlen($ini_value)-1]);
			$ini_value  = substr($ini_value, 0, -1); // necessary since PHP 7.1; otherwise optional

			switch($last) {

				case 'g':
					$ini_value *= 1024;

				case 'm':
					$ini_value *= 1024;

				case 'k':
					$ini_value *= 1024;
			}

			return $ini_value;
		}

		// Get HTTP environment variable (Accepts array for multiple HTTP environment variable checks)
		public static function get_http_env($variable_array) {

			// Checks
			if(!isset($_SERVER)) { return ''; }
			if(!is_array($variable_array)) { $variable_array = array($variable_array); }
			$variable_array_index_last = count($variable_array) - 1;

			// Run through each variable
			foreach($variable_array as $variable_array_index => $variable) {

				if(isset($_SERVER[$variable])) {

					if($variable_array_index == $variable_array_index_last) {

						return $_SERVER[$variable];

					} else {

						if(!empty($_SERVER[$variable])) return $_SERVER[$variable];
					}
				}
			}

			return '';
		}

		// Get hostname
		public static function get_hostname() {

			$protocol = isset($_SERVER) && isset($_SERVER['HTTPS']) && ($_SERVER['HTTPS'] !== 'off' || $_SERVER['SERVER_PORT'] == 443) ? "https://" : "http://";
			return str_replace($protocol, '', strtolower(get_site_url()));
		}

		// Get object meta value
		public static function get_object_meta_value($object, $meta_key, $default_value = '') {

			// Check for meta data
			if(!isset($object->meta)) { return $default_value; }

			// Check for meta_key
			if(!isset($object->meta->{$meta_key})) { return $default_value; }

			return $object->meta->{$meta_key};
		}

		// Extract numbers from string
		public static function get_tel($phone) {

			return preg_replace('/[^+\d]+/', '', $phone);
		}

		// Get array of MIME types
		public static function get_mime_array($value) {

			if(is_array($value)) { return $value; }

			$mime_array = explode(',', $value);
			$mime_array_return = array();

			foreach($mime_array as $mime) {

				$mime_split = $mime_array = explode('/', $mime);
				if(count($mime_split) !== 2) { continue; }
				if(strlen($mime_split[0]) == 0) { continue; }
				if(strlen($mime_split[1]) == 0) { continue; }
				$mime_array_return[] = strtolower(trim($mime));
			}

			return $mime_array_return;
		}

		// Add datestamp to filename
		public static function filename_datestamp($filename_prefix, $filename_suffix) {

			$filename = $filename_prefix . current_time('-Y-m-d-H-i-s') . '.' . $filename_suffix;
			return sanitize_file_name($filename);
		}

		// Output file download headers
		public static function file_download_headers($filename, $mime_type, $encoding = 'binary') {

			$filename = sanitize_file_name($filename);			// WordPress function

			// HTTP headers
			header('Pragma: public');
			header('Expires: 0');
			header('Cache-Control: must-revalidate, post-check=0, pre-check=0');
			header('Cache-Control: private', false);
			header('Content-Type: ' . $mime_type);
			header('Content-Disposition: attachment; filename=' . $filename);
			header('Content-Transfer-Encoding: ' . $encoding);
		}

		public static function hex_to_hsl($hex) {

			$hex = array($hex[0].$hex[1], $hex[2].$hex[3], $hex[4].$hex[5]);
			$rgb = array_map(function($part) {

				return hexdec($part) / 255;

			}, $hex);

			$max = max($rgb);
			$min = min($rgb);

			$l = ($max + $min) / 2;

			if ($max == $min) {

				$h = $s = 0;

			} else {

				$diff = $max - $min;
				$s = $l > 0.5 ? $diff / (2 - $max - $min) : $diff / ($max + $min);

				switch($max) {

					case $rgb[0]:
						$h = ($rgb[1] - $rgb[2]) / $diff + ($rgb[1] < $rgb[2] ? 6 : 0);
						break;

					case $rgb[1]:
						$h = ($rgb[2] - $rgb[0]) / $diff + 2;
						break;

					case $rgb[2]:
						$h = ($rgb[0] - $rgb[1]) / $diff + 4;
						break;
				}

				$h /= 6;
			}

			return array($h, $s, $l);
		}

		public static function hsl_to_rgb($hsl) {

			list($h, $s, $l) = $hsl;

		    $r; 
		    $g; 
		    $b;

			$c = ( 1 - abs( 2 * $l - 1 ) ) * $s;
			$x = $c * ( 1 - abs( fmod( ( $h / 60 ), 2 ) - 1 ) );
			$m = $l - ( $c / 2 );
			if ( $h < 60 ) {
				$r = $c;
				$g = $x;
				$b = 0;
			} else if ( $h < 120 ) {
				$r = $x;
				$g = $c;
				$b = 0;			
			} else if ( $h < 180 ) {
				$r = 0;
				$g = $c;
				$b = $x;					
			} else if ( $h < 240 ) {
				$r = 0;
				$g = $x;
				$b = $c;
			} else if ( $h < 300 ) {
				$r = $x;
				$g = 0;
				$b = $c;
			} else {
				$r = $c;
				$g = 0;
				$b = $x;
			}
			$r = ( $r + $m ) * 255;
			$g = ( $g + $m ) * 255;
			$b = ( $b + $m  ) * 255;

			return array( floor( $r ), floor( $g ), floor( $b ) );
		}

		public static function hsl_to_hex($hsl) {

			list($h, $s, $l) = $hsl;

			if ($s == 0) {

				$r = $g = $b = 1;

			} else {

				$q = $l < 0.5 ? $l * (1 + $s) : $l + $s - $l * $s;
				$p = 2 * $l - $q;

				$r = self::hue_to_rgb($p, $q, $h + 1/3);
				$g = self::hue_to_rgb($p, $q, $h);
				$b = self::hue_to_rgb($p, $q, $h - 1/3);
			}

			return self::rgb_to_hex($r) . self::rgb_to_hex($g) . self::rgb_to_hex($b);
		}

		public static function hue_to_rgb($p, $q, $t) {

			if ($t < 0) $t += 1;
			if ($t > 1) $t -= 1;
			if ($t < 1/6) return $p + ($q - $p) * 6 * $t;
			if ($t < 1/2) return $q;
			if ($t < 2/3) return $p + ($q - $p) * (2/3 - $t) * 6;

			return $p;
		}

		public static function rgb_to_hex($rgb) {

			return str_pad(dechex($rgb * 255), 2, '0', STR_PAD_LEFT);
		}

		public static function hex_hsl_adjust($hex, $hPercent, $sPercent, $lPercent) {

			if($hex == '') { return ''; }

			// Check for hash
			$has_hash = (substr($hex, 0, 1) == '#');
			if($has_hash) { $hex = substr($hex, 1); }

			// Convert to HSL
			$hsl = self::hex_to_hsl($hex);

			$h = $hsl[0];
			$s = $hsl[1];
			$l = $hsl[2];

			// Adjust
			$h = $h + (($hPercent / 100) * $h);
			$s = $s + (($sPercent / 100) * $s);
			$l = $l + (($lPercent / 100) * $l);

			// Convert back to hex
			return ($has_hash ? '#' : '') . self::hsl_to_hex(array($h, $s, $l));
		}

		// Green (0) --> Lime Green (25) --> Yellow (50) --> Orange (75) --> Red (100)
		public static function get_green_to_red_rgb($value, $min = 0, $max = 100) {

			// Calculate ratio
			$ratio = $value / $max;
			if($ratio < 0) { $ratio = 0; }
			if($ratio > 1) { $ratio = 1; }

			// Red
			$r = ($ratio * 2) * 255;
			$r = ($r > 255) ? 255 : $r;

			// Green
			$g = (2 - ($ratio * 2)) * 255;
			$g = ($g > 255) ? 255 : $g;

			// Blue
			$b = 0;

			return "rgb($r,$g,$b)";
		}

		// Parse WS Form variables
		public static function parse_variables_process($parse_string, $form = false, $submit = false, $content_type = 'text/html', $depth = 1) {

			if(!is_string($parse_string)) { return $parse_string; }

			// Checks to speed up this function
			if(strpos($parse_string, '#') === false) { return $parse_string; }

			// Initialize variables
			$variables = [];

			// Initialize meta data
			$form_meta = self::parse_variables_form_meta($form);
			$submit_meta = self::parse_variables_submit_meta($submit);

			// Check for too many iterations
			if($depth > 100) { return ''; }

			// User
			$current_user = WS_Form_Action::get_user();
			if($current_user === false) {

				if(!function_exists('wp_get_current_user')) {

					include_once(ABSPATH . 'wp-includes/pluggable.php');
				}
				$current_user = wp_get_current_user();
			}

			$parse_variables_config = WS_Form_Config::get_parse_variables();

			// Process each parse variable key
			foreach($parse_variables_config as $parse_variables_key => $parse_variables) {

				if(strpos($parse_string, '#' . $parse_variables_key) === false) { continue; }

				foreach($parse_variables['variables'] as $parse_variable => $parse_variable_config) {

					if(strpos($parse_string, '#' . $parse_variable) === false ) { continue; }

					$parsed_variable = '';

					switch($parse_variable) {

						case 'email_tracking' :

							if(!isset($submit->meta)) { break; }

							// Tracking
							$tracking_array = WS_Form_Config::get_tracking(false);

							foreach($tracking_array as $meta_key => $tracking) {

								// Check meta_key is enabled
								if(($form === false) || !WS_Form_Common::get_object_meta_value($form, $meta_key, false)) { break; }

								$label = $tracking['label'];

								// Add to fields_all HTML
								switch($content_type) {

									case 'text/html' :

										$parsed_variable .= '<p style="font-family: sans-serif; font-size: 14px; font-weight: normal; margin: 0; margin-bottom: 10px;"><strong>' . htmlentities($label) . '</strong><br />#' . $meta_key . '</p>';
										break;

									default :

										$parsed_variable .= $label . "\n" . $value . "\n\n";
										break;
								}
							}

							$parsed_variable = self::parse_variables_process($parsed_variable, $form, $submit, $content_type, $depth + 1);

							$variables[$parse_variable] = $parsed_variable;

							break;
						
						default :

							// Assign value
							$parse_variable_value = (isset($parse_variable_config['value'])) ? $parse_variable_config['value'] : '';
							$parse_variable_attributes = (isset($parse_variable_config['attributes'])) ? $parse_variable_config['attributes'] : false;

							// If no attributes specified, then just set the value
							if(($parse_variable_attributes === false) && ($parse_variable_value !== false)) { $variables[$parse_variable] = $parse_variable_value; continue; }

							// Get number of attributes required
							$variable_attribute_count = isset($parse_variable_config['attributes']) ? count($parse_variable_attributes) : 0;

							// Handle variables
							if($variable_attribute_count > 0) {

								// Do until no more found
								$variable_index_start = 0;
								do {

									// Find position of variable and brackets
									$variable_index_of = strpos($parse_string, '#' . $parse_variable, $variable_index_start);

									// No more instances of variable found
									if($variable_index_of === false) { continue; }

									$variable_index_of_bracket_start = strpos($parse_string, '(', $variable_index_of + strlen($parse_variable));
									$variable_index_of_bracket_finish = strpos($parse_string, ')', $variable_index_of_bracket_start);

									// Check brackets found
									if(	($variable_index_of_bracket_start === false) ||
										($variable_index_of_bracket_finish === false) ) {

										// Shift index to look for next instance
										$variable_index_start += strlen('#' . $parse_variable);

										// Get full string to parse
										$parse_variable_full = '#' . $parse_variable;

										// No brackets found so set attributes as blank
										$variable_attribute_array_raw = [];

									} else {

										// Shift index to look for next instance
										$variable_index_start = $variable_index_of_bracket_finish;

										// Ensure bracket starts immediate after parse_variable
										if((($variable_index_of_bracket_start - $variable_index_of) - strlen('#' . $parse_variable)) !== 0) { continue; };

										// Get full string to parse
										$parse_variable_full = substr($parse_string, $variable_index_of, ($variable_index_of_bracket_finish + 1) - $variable_index_of);

										// Get attribute string
										$variable_attribute_string = substr($parse_string, $variable_index_of_bracket_start + 1, ($variable_index_of_bracket_finish - 1) - $variable_index_of_bracket_start);

										// Get attribute array
										$variable_attribute_array = str_getcsv($variable_attribute_string, ',');

										// Trim and strip double quotes
										foreach($variable_attribute_array as $key => $e) {

											$e = trim($e);
											$e = preg_replace('/^"(.+(?="$))"$/', '', $e);
											$variable_attribute_array[$key] = $e;
										}
									}

									// Check each attribute
									foreach($parse_variable_attributes as $parse_variable_attributes_index => $parse_variable_attribute) {

										$parse_variable_attribute_id = $parse_variable_attribute['id'];
										$parse_variable_attribute_valid = isset($parse_variable_attribute['valid']) ? $parse_variable_attribute['valid'] : false;

										// Was attribute provided for this index?
										$parse_variable_attribute_supplied = isset($variable_attribute_array[$parse_variable_attributes_index]);

										// Check required
										$parse_variable_attribute_required = (isset($parse_variable_attribute['required']) ? $parse_variable_attribute['required'] : true);
										if($parse_variable_attribute_required && !$parse_variable_attribute_supplied) {

											// Syntax error - Attribute count
											self::throw_error(sprintf(__('Syntax error, missing attribute: %s (Expected: %s)', 'ws-form'), '#' . $parse_variable, $parse_variable_attribute_id));
											continue;
										}

										// Check default
										$parse_variable_attribute_default = isset($parse_variable_attribute['default']) ? $parse_variable_attribute['default'] : false;

										if(($parse_variable_attribute_default !== false) && !$parse_variable_attribute_supplied) {

											$variable_attribute_array[$parse_variable_attributes_index] = $parse_variable_attribute_default;
										}

										// Check validity
										if($parse_variable_attribute_valid !== false) {

											if(!in_array($variable_attribute_array[$parse_variable_attributes_index], $parse_variable_attribute_valid)) {

												$parse_variable_attribute_valid = $parse_variable_attribute_default;
											}
										}
									}

									// Process variable
									switch($parse_variable) {

										case 'query_var' :

											$parsed_variable = self::get_query_var($variable_attribute_array[0]);
											if($content_type == 'text/html') { $parsed_variable = htmlentities($parsed_variable); }
											break;

										case 'email_submission' :

											if(!isset($submit->meta)) { break; }

											$render_group_labels = $variable_attribute_array[0];
											$render_section_labels = $variable_attribute_array[1];
											$render_blank_fields = $variable_attribute_array[2];

											$value = self::parse_variables_fields_all((object) $form, $submit, $content_type, $render_group_labels, $render_section_labels, $render_blank_fields);

											$parsed_variable = self::parse_variables_process($value, $form, $submit, $content_type, $depth + 1);

											break;

										case 'field' :

											if(!isset($submit->meta)) { break; }

											if(!is_numeric($variable_attribute_array[0])) { break; }

											$field_id = $variable_attribute_array[0];

											if(!isset($submit->meta[WS_FORM_FIELD_PREFIX . $field_id])) { break; }

											// Get value
											$meta = $submit->meta[WS_FORM_FIELD_PREFIX . $field_id];
											$value = self::parse_variables_meta_value($form, $meta, $content_type);

											$parsed_variable = self::parse_variables_process($value, $form, $submit, $content_type, $depth + 1);
											if($content_type == 'text/html') { $parsed_variable = htmlentities($parsed_variable); }

											break;

										case 'post_meta' :

											if(!isset($post)) {

												$post_id = WS_Form_Common::get_query_var('post_id', 0);
												$post = ($post_id > 0) ? get_post($post_id) : null;
											}
											if($post === null) { break; }

											$post_meta = get_post_meta($post);
											if($post_meta === false) { break; }

											$meta_key = $variable_attribute_array[0];

											if(!isset($post_meta[$meta_key])) { break; }
											if(!isset($post_meta[$meta_key][0])) { break; }

											$parsed_variable = $post_meta[$meta_key][0];

											break;

										case 'acf_repeater_field' :

											global $post;

											if(!isset($post)) {

												$post_id = WS_Form_Common::get_query_var('post_id', 0);
												$post = ($post_id > 0) ? get_post($post_id) : null;
											}
											if($post === null) { break; }

											$parent_field = $variable_attribute_array[0];
											$sub_field = $variable_attribute_array[1];

											$parent_field_array = explode(',', $parent_field);
											foreach($parent_field_array as $key => $value) {

												$parent_field_array[$key] = trim($value);
											}

											$parsed_variable = self::acf_repeater_field_walker($parent_field_array, $sub_field);

											break;

										// Date
										case 'post_date_custom' :
										case 'server_date_custom' :

											$parsed_variable = date($variable_attribute_array[0], strtotime($parse_variable_value));
											if($content_type == 'text/html') { $parsed_variable = htmlentities($parsed_variable); }
											break;

										// User
										case 'user_lost_password_url' :

											// Check we have user data
											if(($current_user === false) || !$current_user->ID) { break; }

											// Check we can produce a lost password URL
											if(!(

												isset($current_user->lost_password_key) && 
												($current_user->lost_password_key != '') && 
												isset($current_user->user_login) && 
												($current_user->user_login != '')

											)) { break; }

											// Get path
											$path = $variable_attribute_array[0];

											if($path !== '') {

												$parsed_variable = network_site_url(sprintf('%s?key=%s&login=%s', $path, rawurlencode($current_user->lost_password_key),rawurlencode($current_user->user_login)));

											} else {

												$parsed_variable = network_site_url(sprintf('wp-login.php?action=rp&key=%s&login=%s', rawurlencode($current_user->lost_password_key), rawurlencode($current_user->user_login)), 'login');
											}

											break;
									}

									// Assign value
									$variables[substr($parse_variable_full, 1)] = $parsed_variable;

								} while ($variable_index_of !== false);
							}
					}
				}
			}

			// Form
			if(strpos($parse_string, 'form')) {

				$variables['form_label'] = $form_meta['label'];
				$variables['form_id'] = $form_meta['id'];
				$variables['form_checksum'] = $form_meta['published_checksum'];

				// These variables are only available on the public side
				$variables['form_obj_id'] = '';
				$variables['form_framework'] = '';
				$variables['form_instance_id'] = 0;
			}

			// Submit
			if(strpos($parse_string, 'submit')) {

				$variables['submit_id'] = $submit->id;
				$variables['submit_user_id'] = $submit_meta['user_id'];
				$variables['submit_hash'] = $submit_meta['hash'];
			}

			// WS Form
			if(strpos($parse_string, 'email')) {

				$variables['email_promo'] = sprintf(__('Powered by %s.', 'ws-form'), sprintf(($content_type == 'text/html') ? '<a href="%s" style="color: #999999; font-size: 12px; text-align: center; text-decoration: none;">WS Form</a>' : 'WS Form %s', self::get_plugin_website_url('', 'email_footer')));
			}

			// Tracking
			if(strpos($parse_string, 'tracking')) {

				$tracking_array = WS_Form_Config::get_tracking(false);

				foreach($tracking_array as $meta_key => $tracking) {

					// Check tracking meta_key is enabled
					if(($form === false) || !WS_Form_Common::get_object_meta_value($form, $meta_key, false)) { break; }

					// Check tracking meta_key exists in parse string
					if(strpos($parse_string, '#' . $meta_key) === false) { break; }

					// Check meta exists in submit data
					if(!isset($submit->meta)) { break; }
					if(!isset($submit->meta[$meta_key])) {

						$value = '';

					} else {

						// Get value
						$meta = array(

							'type'	=>	$tracking['type'],
							'value'	=>	$submit->meta[$meta_key]
						);

						// Get parsed version
						$value = self::parse_variables_meta_value($form, $meta, $content_type);
					}

					$variables[$meta_key] = $value;
				}
			}

			// Parse
			$parse_string = self::mask_parse($parse_string, $variables);

			// Filter
			$parse_string = apply_filters('wsf_config_parse_string', $parse_string);

			return $parse_string;
		}

		// ACF parent field crawler
		public static function acf_repeater_field_walker($parent_field_array, $sub_field) {

			$return_value = '';

			$parent_field = array_shift($parent_field_array);

			if(have_rows($parent_field)) {

				while(have_rows($parent_field)) {

					the_row();

					$row = get_row();

					if(count($parent_field_array) == 0) {

						$sub_field_value = get_sub_field($sub_field);

						if($sub_field_value !== false) { return $sub_field_value; }

					} else {

						$return_value = self::acf_repeater_field_walker($parent_field_array, $sub_field);
					}
				}
			}

			return $return_value;
		}

		// Parse form data for use with parse_variables
		public static function parse_variables_meta_value($form, $meta, $content_type) {

			$type = $meta['type'];
			$value = $meta['value'];

			// HTML encode values
			if($content_type == 'text/html' && !is_array($value)) {

				$value = htmlentities($value);

				switch($type) {

					case 'url' :

						$value = sprintf('<a href="%1$s" target="_blank">%1$s</a>', $value);
						break;

					case 'tel' :

						$value = sprintf('<a href="tel:%1$s">%1$s</a>', $value);
						break;

					case 'email' :

						if(filter_var($value, FILTER_VALIDATE_EMAIL)) {

							$value = sprintf('<a href="mailto:%1$s">%1$s</a>', $value);
						}
						break;

					case 'ip' :

						// Get lookup URL mask
						$ip_lookup_url_mask = WS_Form_Common::option_get('ip_lookup_url_mask');
						if(empty($ip_lookup_url_mask)) { $value = htmlentities($value); break; }

						// Get #value for mask
						$ip_lookup_url_mask_values = array('value' => $value);

						// Build lookup URL
						$ip_lookup_url = self::mask_parse($ip_lookup_url_mask, $ip_lookup_url_mask_values);

						$value = '<a href="' . $ip_lookup_url . '" target="_blank">' . htmlentities($value) . '</a>';
						break;

					case 'latlon' :

						if(preg_match('/^(\-?\d+(\.\d+)?),\s*(\-?\d+(\.\d+)?)$/', $value) == 1) {

							// Get lookup URL mask
							$latlon_lookup_url_mask = WS_Form_Common::option_get('latlon_lookup_url_mask');
							if(empty($latlon_lookup_url_mask)) { $value = htmlentities($value); break; }

							// Get #value for mask
							$latlon_lookup_url_mask_values = array('value' => $value);

							// Build lookup URL
							$latlon_lookup_url = self::mask_parse($latlon_lookup_url_mask, $latlon_lookup_url_mask_values);

							$value = '<a href="' . $latlon_lookup_url . '" target="_blank">' . htmlentities($value) . '</a>';

						} else {

							switch(intval($value)) {

								case 1 :

									$value = __('User denied the request for geo location', 'ws-form');
									break;

								case 2 :

									$value = __('Geo location information was unavailable', 'ws-form');
									break;

								case 3 :

									$value = __('The request to get user geo location timed out', 'ws-form');
									break;

								default :

									$value = '-';
							}
						}
						break;
				}
			}

			// Process by field type
			switch($type) {

				case 'file' :
				case 'signature' :

					$files = $meta['value'];

					if(empty($files)) { break; }

					$value_array = array();

					foreach($files as $file) {

						$file_name = $file['name'];
						$file_size = self::get_file_size($file['size']);

						$value_array[] = $file_name . ' (' . $file_size . ')';
					}

					$value = implode((($content_type == 'text/html') ? '<br />' : "\n"), $value_array);

					break;

				case 'datetime' :

					$fields = WS_Form_Common::get_fields_from_form($form);

					if(!isset($fields[$meta['id']])) { break; }

					$field = $fields[$meta['id']];

					$input_type_datetime = WS_Form_Common::get_object_meta_value((object)$field, 'input_type_datetime', 'date');

					$value = WS_Form_Common::get_date_by_type($value, $input_type_datetime);

					break;

				default :

					$value = is_array($value) ? implode(', ', $value) : $value;
			}

			return $value;
		}

		// Parse form data for use with parse_variables
		public static function parse_variables_form_meta($form) {

			if($form === false) { return array(); }

			$form_meta = array(

				'id'					=>	isset($form->id) ? $form->id : '',
				'label'					=>	isset($form->label) ? $form->label : '',
				'published_checksum'	=>	isset($form->published_checksum) ? $form->published_checksum : '',
			);

			return $form_meta;
		}

		// Parse submit data for use with parse_variables
		public static function parse_variables_submit_meta($submit) {

			if($submit === false) { return array(); }

			$submit_meta = array(

				'hash'		=>	isset($submit->hash) ? $submit->hash : '',
				'user_id'	=>	isset($submit->user_id) ? $submit->user_id : '',
			);

			return $submit_meta;
		}

		// Add populate values to fields
		public static function parse_variables_fields_all($form, $submit, $content_type, $render_group_labels, $render_section_labels, $render_blank_fields) {

			$fields_all = self::parse_variables_fields_all_group($form->groups, $form, $submit, $content_type, $render_group_labels, $render_section_labels, $render_blank_fields);

			return $fields_all;
		}

		// Run through each group
		public static function parse_variables_fields_all_group($groups, $form, $submit, $content_type, $render_group_labels, $render_section_labels, $render_blank_fields) {

			$fields_all = '';

			$group_count = count($groups);
			$group_label_join = '';

			foreach($groups as $key => $group) {

				// Should label be rendered?
				$render_label =	($render_group_labels == 'true') ||
								(($render_group_labels == 'auto') && self::get_object_meta_value($group, 'label_render'));

				if(isset($groups[$key]->sections)) {

					$section_fields = self::parse_variables_fields_all_section($group->sections, $form, $submit, $content_type, $render_section_labels, $render_blank_fields);

					if(($group_count > 0) && ($section_fields != '') && $render_group_labels) {

						switch($content_type) {

							case 'text/html' :

								$fields_all .= $group_label_join . ($render_label ? '<h2 style="font-family: sans-serif; font-size: 24px; font-weight: bold; margin: 0; margin-bottom: 10px;">' . htmlentities($group->label) . '</h2>' : '');
								$group_label_join = '<hr style="margin: 20px 0" />';
								break;

							default :

								$fields_all .= $group_label_join . "** " . $group->label . " **\n\n";
								$group_label_join = "\n";
						}

					}

					$fields_all .= $section_fields;
				}
			}

			return $fields_all;
		}

		// Run through each section
		public static function parse_variables_fields_all_section($sections, $form, $submit, $content_type, $render_section_labels, $render_blank_fields) {

			$fields_all = '';

			$section_count = count($sections);
			$section_label_join = '';

			// Get field types in single dimension array
			$field_types = WS_Form_Config::get_field_types_flat();

			foreach($sections as $key => $section) {

				// Should label be rendered?
				$render_label =	($render_section_labels == 'true') ||
								($render_section_labels == 'auto' && self::get_object_meta_value($section, 'label_render'));

				if(isset($sections[$key]->children)) {

					$fields_all .= self::parse_variables_fields_all_section($section->children, $form, $submit, $content_type, $render_section_labels, $render_blank_fields);
				}

				if(isset($sections[$key]->fields)) {

					$section_fields = array();

					if(count($sections[$key]->fields) && $render_section_labels) {

						switch($content_type) {

							case 'text/html' :

								$fields_all .= $render_label ? '<h3 style="font-family: sans-serif; font-size: 20px; font-weight: bold; margin: 0; margin-bottom: 10px;">' . htmlentities($section->label) . '</h3>' : '';
								break;

							default :

								$fields_all .= $section_label_join . "* " . $section->label . " *\n\n";
								$section_label_join = "\n\n";
						}
					}

					foreach($section->fields as $field) {

						$type = $field->type;
						$field_type = $field_types[$type];
						$submit_save = isset($field_type['submit_save']) ? $field_type['submit_save'] : false;
						if(!$submit_save) { continue; }

						// Get field label
						$label = $field->label;

						if(isset($submit->meta[WS_FORM_FIELD_PREFIX . $field->id])) {

							// Get submit meta
							$meta = $submit->meta[WS_FORM_FIELD_PREFIX . $field->id];

							// Get field value
							$value = self::parse_variables_meta_value($form, $meta, $content_type);
						}

						// No submit value found
						if($value == '') {

							if($render_blank_fields) {

								$value = '-';

							} else {

								continue;
							}
						}

						// Add to fields_all HTML
						switch($content_type) {

							case 'text/html' :

								$fields_all .= '<p style="font-family: sans-serif; font-size: 14px; font-weight: normal; margin: 0; margin-bottom: 10px;"><strong>' . htmlentities($label) . '</strong><br />' . $value . '</p>';
								break;

							default :

								$fields_all .= $label . "\n" . $value . "\n\n";
								break;
						}
					}
				}
			}

			return $fields_all;
		}

		// Check if user can do a WordPress capability (current_user_can not available on public side)
		public static function can_user($capability) {

			if(!function_exists('wp_get_current_user')) {

				include_once(ABSPATH . 'wp-includes/pluggable.php');
			}

			return current_user_can($capability);
		}

		// Loader
		public static function loader() {
?>
<!-- Loader -->
<div id="wsf-loader"><svg viewBox="25 25 50 50"><circle cx="50" cy="50" r="20" fill="none" stroke-width="5" stroke-miterlimit="10"/></svg></div>
<!-- /Loader -->
<?php
		}

		// Check edition
		public static function is_edition($edition) {

			switch($edition) {

				case 'basic' :

					return true;

				case 'pro' :

					return false;

				default :

					return false;
			}
		}

		// Build data grid meta
		public static function build_data_grid_meta($meta_key, $group_name = false, $columns = false, $rows = false) {

			// Get base meta
			$meta_keys = WS_Form_Config::get_meta_keys();
			if(!isset($meta_keys[$meta_key])) { return false; }
			if(!isset($meta_keys[$meta_key]['default'])) { return false; }
			$meta = $meta_keys[$meta_key]['default'];

			if($group_name !== false) { $meta['groups'][0]['label'] = $group_name; }
			if($columns !== false) { $meta['columns'] = $columns; }
			if($rows !== false) { $meta['groups'][0]['rows'] = $rows; }

			return $meta;
		}

		// Get nice file size
		public static function get_file_size($bytes) {

			if($bytes >= 1048576) {

				$bytes = number_format($bytes / 1048576, 2) . ' MB';

			} elseif ($bytes >= 1024) {

				$bytes = number_format($bytes / 1024, 2) . ' KB';

			} elseif ($bytes > 1) {

				$bytes = $bytes . ' bytes';

			} elseif ($bytes == 1) {

				$bytes = $bytes . ' byte';

			} else {

				$bytes = '0 bytes';
			}

			return $bytes;
		}

		// Get nice date by type
		public static function get_date_by_type($date, $type) {

			if(empty($date)) { return ''; }

			switch($type) {

				case 'date' :

					return date(get_option('date_format'), strtotime($date));

				case 'month' :

					return date('F Y', strtotime($date));

				case 'time' :

					return date(get_option('time_format'), strtotime($date));

				case 'week' :

					return __('Week', 'ws-form') . ' ' . date('W, Y', strtotime($date));

				default :

					return date(get_option('date_format') . ' ' . get_option('time_format'), strtotime($date));
			}
		}

		// Throw error
		public static function throw_error($error) {
			
			throw new Exception($error);
		}

		// Get system report
		public static function get_system_report_html() {

			// Get system report
			$system_report = WS_Form_Config::get_system();

			// Build system report HTML
			$system_report_html = '<table class="wsf-table-system">';

			foreach($system_report as $group_id => $group) {

				$system_report_html .= '<tbody>';

				$system_report_html .= '<tr><th colspan="2"><h2>' . htmlentities($group['label']) . '</h2></th></tr>';

				foreach($group['variables'] as $item_id => $item) {

					// Valid
					// 0 = Ignore, 1 = Yes, 2 = No
					$valid = isset($item['valid']) ? ($item['valid'] ? 1 : 2) : 0;

					$system_report_html .= '<tr';

					switch($valid) {

						case 1 : $system_report_html .= ' class="wsf-system-valid"'; break;
						case 2 : $system_report_html .= ' class="wsf-system-invalid"'; break;
					}

					// Label
					$system_report_html .= '><td><b>' . htmlentities($item['label']);
					if(isset($item['min'])) { $system_report_html .= ' (Min: ' . $item['min'] . ')'; }
					$system_report_html .= '</b></td>';

					// Value
					$system_report_html .= '<td>';

					$value = isset($item['value']) ? $item['value'] : '-';
					$type = isset($item['type']) ? $item['type'] : 'text';

					switch($type) {

						case 'plugins' :

							if(is_array($value)) {

								$plugin_array = array();

								foreach($value as $plugin_path) {

									$plugin_data = get_plugin_data(WP_PLUGIN_DIR . '/' . $plugin_path);
									$plugin_array[] = sprintf('<a href="%s" target="_blank">%s</a> (%s)', $plugin_data['PluginURI'], $plugin_data['Name'], $plugin_data['Version']);
								}

								$value = implode('<br />', $plugin_array);
							}
							break;

						case 'boolean' :

							$value = $value ? __('Yes', 'ws-form') : __('No', 'ws-form');
							break;

						case 'url' :

							$value = sprintf('<a href="%1$s" target="_blank">%1$s</a>', $value);
							break;

						case 'size' :

							$value = size_format($value);
							break;

						case 'edition' :

							switch($value) {

								case 'basic' : $value = 'Basic'; break;
								case 'pro' : $value = 'Pro'; break;
							}
							break;

						case 'date' :

							$value = ($value != '') ? date(get_option('date_format'), $value) : '-';
							break;
					}

					$system_report_html .= $value;

					// Suffix
					if(isset($item['suffix'])) { $system_report_html .= ' ' . $item['suffix']; }

					// Valid
					switch($valid) {

						case 1 : $system_report_html .= WS_Form_Config::get_icon_16_svg('check'); break;
						case 2 : $system_report_html .= WS_Form_Config::get_icon_16_svg('warning'); break;
					}

					$system_report_html .= '</td></tr>';
				}

				$system_report_html .= '</tbody>';
			}

			$system_report_html .= '</table>';

			return $system_report_html;
		}

		// Get a string formatted for SMTP email addresses
		public static function get_email_address($email, $name = '') {

			// Ensure email is valid
			if(!filter_var($email, FILTER_VALIDATE_EMAIL)) { return false; }

			// Escape double quotes in name
			$name = str_replace('"', '\"', $name);
			if(strlen($name) > 255) { $name = ''; }

			// Return full email address
			return (($name != '') ? ('"' . $name . '" ') : '') . '<' . $email . '>';
		}
	}
