<?php

	// WS Form Widget

	class WS_Form_Widget extends WP_Widget {

		// Main constructor
		public function __construct() {

			parent::__construct(

				'ws_form_widget',
				WS_FORM_NAME_PRESENTABLE,

				array(
					'description' => __('Displays a form from your WS Form form library.', 'ws-form'),
					'customize_selective_refresh' => true,
				)
			);
		}

		// The widget form (for the backend )
		public function form( $instance ) {	

			// Set widget defaults
			$defaults = array(

				'form_id'    => '',
			);
	
			// Parse current settings with defaults
			extract( wp_parse_args( ( array ) $instance, $defaults ) );

			// Get forms from API
			$ws_form_form = New WS_Form_Form();
			$forms = $ws_form_form->db_read_all('', 'NOT status="trash"', 'label', '', '', false);

			if($forms) {
?>
<p><label for="<?php echo esc_attr($this->get_field_id('form_id')); ?>"><?php _e('Select the form you want to add...', 'ws-form' ); ?></label>

<select id="<?php echo esc_attr($this->get_field_id('form_id')); ?>" name="<?php echo esc_attr($this->get_field_name('form_id')); ?>">
<option value=""><?php _e('Select...', 'ws-form'); ?></option>
<?php
				foreach($forms as $form) {

?><option value="<?php echo $form['id']; ?>"<?php if($form['id'] == $form_id) { ?> selected="selected"<?php } ?>><?php echo $form['label']; ?></option>
<?php
				}
?>
</select></p>
<?php
			} else {
?>
<p><?php _e("You haven't created any forms yet.", 'ws-form'); ?></p>
<p><a href="<?php echo WS_Form_Common::get_admin_url('ws-form-add'); ?>"><?php _e('Click here to create a form', 'ws-form'); ?></a></p>
<?php
			}
		}

		// Update widget settings
		public function update($new_instance, $old_instance) {

			$instance = $old_instance;
			$instance['form_id']    = isset($new_instance['form_id'] ) ? wp_strip_all_tags($new_instance['form_id']) : '';
			return $instance;
		}

		// Display the widget
		public function widget($args, $instance) {

			extract($args);

			// Check the widget options
			$form_id = isset($instance['form_id']) ? $instance['form_id'] : '';

			// WordPress core before_widget hook (always include )
			echo $before_widget;

			// Display the widget
			echo do_shortcode('[' . WS_FORM_SHORTCODE . ' id=' . $form_id . ']');

			// WordPress core after_widget hook (always include )
			echo $after_widget;
		}

	}

	// Register the widget
	function ws_form_widget() {

		register_widget('WS_Form_Widget');
	}
	add_action('widgets_init', 'ws_form_widget');
