<?php

	class WS_Form_Submit extends WS_Form_Core {

		public $id;
		public $form_id;
		public $hash;
		public $duration;
		public $count_submit;
		public $meta;
		public $meta_protected;
		public $actions;
		public $preview;
		public $date_added;
		public $date_updated;
		public $date_expire;
		public $spam_level;

		public $event;

		public $error;
		public $error_message;
		public $error_code;

		public $error_validation;
		public $error_messages_validation;

		public $table_name;
		public $table_name_meta;

		const DB_INSERT = 'form_id,date_added,date_updated,date_expire,user_id,hash,status,actions,preview,spam_level';
		const DB_UPDATE = 'form_id,date_updated,user_id,hash,actions';
		const DB_SELECT = 'form_id,date_added,date_updated,date_expire,user_id,hash,duration,status,actions,preview,spam_level,id';

		public function __construct() {

			global $wpdb;

			$this->id = 0;
			$this->form_id = 0;
			$this->table_name = $wpdb->prefix . WS_FORM_DB_TABLE_PREFIX . 'submit';
			$this->table_name_meta = $this->table_name . '_meta';
			$this->hash = '';
			$this->status = 'publish';
			$this->duration = 0;
			$this->count_submit = 0;
			$this->meta = array();
			$this->actions = '';
			$this->preview = false;
			$this->date_added = WS_Form_Common::get_mysql_date();
			$this->date_updated = WS_Form_Common::get_mysql_date();
			$this->date_expire = null;
			$this->spam_level = null;

			$this->event = false;

			$this->error = false;
			$this->error_message = '';
			$this->error_code = 200;

			$this->error_validation = false;
			$this->error_messages_validation = array();
		}

		// Create
		public function db_create() {

			// No capabilities required, this is a public method

			// Check form ID
			self::db_check_form_id();

			global $wpdb;

			// get_user_id(false) = Does not exit on zero
			$sql = sprintf("INSERT INTO %s (%s) VALUES (%u, NOW(), NOW(), %s, %u, '%s', '%s', '%s', %u, %s);", $this->table_name, self::DB_INSERT, $this->form_id, (is_null($this->date_expire) ? 'NULL' : "'" . $this->date_expire . "'"), WS_Form_Common::get_user_id(false), esc_sql($this->hash), 'draft', esc_sql($this->actions), ($this->preview ? 1 : 0), (is_null($this->spam_level) ? 'NULL' : $this->spam_level));
			if($wpdb->query($sql) === false) { self::db_throw_error(__('Error adding submit', 'ws-form')); }

			// Get inserted ID
			$this->id = $wpdb->insert_id;

			// Create hash
			self::db_create_hash();

			// Update hash
			$sql = sprintf("UPDATE %s SET hash = '%s' WHERE id=%u LIMIT 1", $this->table_name, esc_sql($this->hash), $this->id);
			if($wpdb->query($sql) === false) { self::db_throw_error(__('Error updating submit', 'ws-form')); }
		}

		// Read record to array
		public function db_read($get_meta = true, $get_expanded = true) {

			// User capability check
			if(!WS_Form_Common::can_user('read_submission')) { return false; }

			self::db_check_id();

			global $wpdb;

			// Add fields
			$sql = sprintf("SELECT %s FROM %s WHERE id = %s LIMIT 1;", self::DB_SELECT, $this->table_name, $this->id);
			$return_array = $wpdb->get_row($sql, 'ARRAY_A');

			if($return_array === null) { parent::db_throw_error(__('Unable to read submission', 'ws-form')); }

			// Set class variables
			foreach($return_array as $key => $value) {

				$this->{$key} = $value;
			}

			// Process meta data
			if($get_meta) {

				$this->meta = self::db_get_submit_meta($return_array);
				$return_array['meta'] = $this->meta;
			}

			// Get user data
			if($get_expanded) {

				self::db_read_expanded($return_array);
			}

			// Return array
			return $return_array;
		}

		// Read expanded data for a record
		public function db_read_expanded(&$return_array, $expand_user = true, $expand_date_added = true, $expand_date_updated = true, $expand_status = true) {

			// User capability check
			if(!WS_Form_Common::can_user('read_submission')) { return false; }

			if($expand_user && isset($return_array['user_id']) && ($return_array['user_id'] > 0)) {

				$user = get_user_by('ID', $return_array['user_id']);
				if($user !== false) {

					$return_array['user'] = array(

						'first_name' 	=>	$user->first_name,
						'last_name' 	=>	$user->last_name,
						'display_name'	=> $user->display_name
					);
				}
			}

			// Date added
			if($expand_date_added && isset($return_array['date_added'])) {

				$return_array['date_added_wp'] = date_i18n(get_option('date_format') . ' ' . get_option('time_format'), strtotime(get_date_from_gmt($return_array['date_added'])));
			}

			// Date updated
			if($expand_date_updated && isset($return_array['date_updated'])) {

				$return_array['date_updated_wp'] = date_i18n(get_option('date_format') . ' ' . get_option('time_format'), strtotime(get_date_from_gmt($return_array['date_updated'])));
			}

			// Status
			if($expand_status && isset($return_array['status'])) {

				$return_array['status_full'] = self::db_get_status_name($return_array['status']);
			}

			// Unserialize action fields
			$return_array['actions'] = (@unserialize($return_array['actions']) !== false) ? unserialize($return_array['actions']) : false;
		}

		// Read - All
		public function db_read_all($join = '', $where = '', $order_by = '', $limit = '', $offset = '', $get_meta = true, $get_expanded = true) {

			// User capability check
			if(!WS_Form_Common::can_user('read_submission')) { return false; }

			$field_cache = array();

			global $wpdb;

			// Get form data
			$select = self::DB_SELECT;
			if($join != '') {

				$select_array = explode(',', $select);
				foreach($select_array as $key => $select) {

					$select_array[$key] = $this->table_name . '.' . $select;
				}
				$select = implode(',', $select_array);
			}

			$sql = sprintf("SELECT %s FROM %s", $select, $this->table_name);

			if($join != '') { $sql .= sprintf(" %s", $join); }
			if($where != '') { $sql .= sprintf(" WHERE %s", $where); }
			if($order_by != '') { $sql .= sprintf(" ORDER BY %s", $order_by); }
			if($limit != '') { $sql .= sprintf(" LIMIT %s", $limit); }
			if($offset != '') { $sql .= sprintf(" OFFSET %s", $offset); }

			$return_array = $wpdb->get_results($sql, 'ARRAY_A');

			// Process meta data
			if($return_array && $get_meta) {

				foreach($return_array as $key => $submit) {

					// Get meta data
					$submit_meta = self::db_get_submit_meta($submit);

					// Add meta to return array
					$return_array[$key] = $return_array[$key] + $submit_meta;
				}
			}

			// Process expanded data
			if($get_expanded) {

				foreach($return_array as $key => $submit) {

					self::db_read_expanded($submit);
					$return_array[$key] = $submit;
				}				
			}

			return $return_array;
		}

		// Read - Count
		public function db_read_count($join = '', $where = '') {

			// User capability check
			if(!WS_Form_Common::can_user('read_submission')) { return false; }

			global $wpdb;

			// Get form data
			$select = self::DB_SELECT;
			if($join != '') {

				$select_array = explode(',', $select);
				foreach($select_array as $key => $select) {

					$select_array[$key] = $this->table_name . '.' . $select;
				}
				$select = implode(',', $select_array);
			}

			$sql = sprintf("SELECT COUNT(id) FROM %s", $this->table_name);

			if($join != '') { $sql .= sprintf(" %s", $join); }
			if($where != '') { $sql .= sprintf(" WHERE %s", $where); }

			return $wpdb->get_var($sql);
		}

		// Read by hash
		public function db_read_by_hash($get_meta = true, $get_expanded = true, $form_id_check = true) {

			// User capability check
			if(!WS_Form_Common::can_user('read_submission')) { return false; }

			// Check form ID
			if($form_id_check) { self::db_check_form_id(); }

			// Check hash
			self::db_check_hash();

			global $wpdb;

			// Get form submission
			if($form_id_check) {

				$sql = sprintf("SELECT %s FROM %s WHERE form_id = %u AND hash = '%s' LIMIT 1;", self::DB_SELECT, $this->table_name, $this->form_id, $this->hash);
			} else {

				$sql = sprintf("SELECT %s FROM %s WHERE hash = '%s' LIMIT 1;", self::DB_SELECT, $this->table_name, $this->hash);				
			}
			$return_array = $wpdb->get_row($sql, 'ARRAY_A');
			if($return_array === null) { $this->hash = ''; return false; }

			// Set class variables
			foreach($return_array as $key => $value) {

				$this->{$key} = $value;
			}

			// Process meta data
			if($get_meta) {

				$return_array['meta'] = self::db_get_submit_meta($return_array);
			}

			// Get user data
			if($get_expanded) {

				self::db_read_expanded($return_array);
			}

			// Return array
			return $return_array;
		}

		// Update current submit
		public function db_update() {

			// No capabilities required, this is a public method

			// Check ID
			self::db_check_id();

			// Update / Insert
			$this->id = parent::db_update_insert($this->table_name, self::DB_UPDATE, self::DB_INSERT, (array)$this, 'submit', $this->id);

			// Update meta
			if(isset($this->meta)) {

				$ws_form_submit_meta = New WS_Form_Submit_Meta();
				$ws_form_submit_meta->parent_id = $this->id;
				$ws_form_submit_meta->db_update_from_object($this->meta);
			}
		}

		// Push submit from array
		public function db_update_from_object($submit) {

			// No capabilities required, this is a public method

			// Check for submit ID in $submit
			if(isset($submit['id'])) { $this->id = absint($submit['id']); } else { return false; }

			// Update / Insert
			$this->id = parent::db_update_insert($this->table_name, self::DB_UPDATE, self::DB_INSERT, $submit, 'submit', $this->id);

			// Update meta
			if(isset($submit['meta'])) {

				$ws_form_submit_meta = New WS_Form_Submit_Meta();
				$ws_form_submit_meta->parent_id = $this->id;
				$ws_form_submit_meta->db_update_from_array($submit['meta']);
			}
		}

		// Stamp submit with date updated, increase submit count and add duration (if available)
		public function stamp() {

			// No capabilities required, this is a public method

			// Check ID
			self::db_check_id();

			// Get duration
			$this->duration = intval(WS_Form_Common::get_query_var('duration', 0));

			global $wpdb;

			// Date updated, count submit + 1
			$sql = sprintf("UPDATE %s SET date_updated = NOW(), count_submit = count_submit + 1, duration = %u WHERE id=%u LIMIT 1", $this->table_name, $this->duration, $this->id);
			if($wpdb->query($sql) === false) { parent::error(__('Error updating submit date updated', 'ws-form')); }
			$this->count_submit++;

			// User ID
			$sql = sprintf("UPDATE %s SET user_id = %u WHERE id=%u AND (user_id = 0 OR user_id IS NULL) LIMIT 1", $this->table_name, WS_Form_Common::get_user_id(false), $this->id);
			if($wpdb->query($sql) === false) { parent::error(__('Error updating submit user ID', 'ws-form')); }
		}

		// Delete
		public function db_delete($permanent_delete = false) {

			// User capability check
			if(!WS_Form_Common::can_user('delete_submission')) { return false; }

			self::db_check_id();

			// Read the submit status
			self::db_read(false, false);

			if($this->status == 'trash') { $permanent_delete = true; }

			// If status is trashed, do a permanent delete of the data
			if($permanent_delete) {

				global $wpdb;

				// Delete form
				$sql = sprintf("DELETE FROM %s WHERE id = %u;", $this->table_name, $this->id);
				if($wpdb->query($sql) === false) { parent::db_throw_error(__('Error deleting submit', 'ws-form')); }

				// Delete meta
				$ws_form_meta = New WS_Form_Submit_Meta();
				$ws_form_meta->parent_id = $this->id;
				$ws_form_meta->db_delete_by_submit();

			} else {

				// Set status to 'trash'
				self::db_set_status('trash');
			}

			return true;
		}

		// Delete trashed submnits
		public function db_trash_delete() {

			// User capability check
			if(!WS_Form_Common::can_user('delete_submission')) { return false; }

			self::db_check_form_id();

			// Get all trashed forms
			$submits = self::db_read_all('', "status='trash' AND form_id=" . $this->form_id, '', '', '', false, false);

       		foreach($submits as $submit) {

				$this->id = $submit['id'];
				self::db_delete();
			}

			return true;
		}

		// Export by email
		public function db_exporter($email_address) {

			global $wpdb;

			// User capability check
			if(!WS_Form_Common::can_user('read_submission')) { return false; }

			// Check email address
			if(!filter_var($email_address, FILTER_VALIDATE_EMAIL)) { return false; }

			$data_to_export = array();

			// Get submit records
			$sql = sprintf('SELECT %1$s.id FROM %2$s LEFT OUTER JOIN %1$s ON %1$s.id = %2$s.parent_id WHERE (LOWER(%2$s.meta_value) = \'%3$s\') AND NOT (%1$s.id IS NULL);', $this->table_name, $this->table_name_meta, esc_sql(strtolower($email_address)));

			$submissions = $wpdb->get_results($sql, 'ARRAY_A');

			// Process results
			if($submissions) {

				foreach($submissions as $submission) {

					// Reset submit data
					$submit_data = array();

					// Get submit ID
					$submit_id = $submission['id'];

					// Get submit record
					$this->id = $submit_id;
					$submit = self::db_read();

					// Remove some data that will not be shared for security reasons or internal only
					unset($submit['form_id']);
					unset($submit['user_id']);
					unset($submit['id']);
					unset($submit['actions']);
					unset($submit['preview']);
					unset($submit['status']);

					// Push all submit data
					foreach($submit as $key => $value) {

						if(!is_array($value)) {

							$submit_data[] = array('name' => $key, 'value' => $value);

						} else {

							switch($key) {

								case 'meta' :

									foreach($value as $meta_key => $meta_value) {

										if(is_array($value)) {

											$value = $meta_value['value'];
											$value = is_array($value) ? print_r($value, true) : $value;

										} else {

											$value = $meta_value;											
										}

										$submit_data[] = array('name' => $meta_key, 'value' => $value);
									}

									break;
							}
						}
					}

					$data_to_export[] = array(
						'group_id'    => WS_FORM_USER_REQUEST_IDENTIFIER,
						'group_label' => __('Form Submissions', 'ws-form'),
						'item_id'     => WS_FORM_USER_REQUEST_IDENTIFIER . '-' . $submit['hash'],
						'data'        => $submit_data
					);
				}
			}

			// Return
			return array(

				'data' => $data_to_export,
				'done' => true,
			);
		}

		// Erase by email
		public function db_eraser($email_address) {

			global $wpdb;

			// User capability check
			if(!WS_Form_Common::can_user('delete_submission')) { return false; }

			// Check email address
			if(!filter_var($email_address, FILTER_VALIDATE_EMAIL)) { return false; }

			// Return array
			$items_removed_count = 0;
			$items_retained_count = 0;

			// Get submit records to be deleted
			$sql = sprintf('SELECT %1$s.id FROM %2$s LEFT OUTER JOIN %1$s ON %1$s.id = %2$s.parent_id WHERE (LOWER(%2$s.meta_value) = \'%3$s\') AND NOT (%1$s.id IS NULL);', $this->table_name, $this->table_name_meta, esc_sql(strtolower($email_address)));

			$submissions = $wpdb->get_results($sql, 'ARRAY_A');

			// Process results
			if($submissions) {

				$items_retained_count = count($submissions);

				if($items_retained_count > 0) {

					// Get first record (Delete one record each time eraser is requested to avoid timeouts)
					if(isset($submissions[0]['id'])) {

						// Delete submit record with permanent delete
						$this->id = $submissions[0]['id'];
						self::db_delete(true);

						$items_retained_count--;
						$items_removed_count++;
					}
				}
			}

			// Build return values
			$items_removed = ($items_removed_count > 0);
			$items_retained = ($items_retained_count > 0);
			$done = ($items_retained <= 0);
			$messages = (($items_removed > 0) && ($items_retained <= 0)) ? array(__('WS Form submissions successfully deleted.', 'ws-form')) : array();

			// Return
			return array(

				'items_removed' => $items_removed,
				'items_retained' => $items_retained,
				'messages' => $messages,
				'done' => $done,
			);
		}

		// Delete expired
		function db_delete_expired() {

			global $wpdb;

			$sql = sprintf("UPDATE %s SET status = 'trash' WHERE (NOT date_expire IS NULL) AND (NOT date_expire = '0000-00-00 00:00:00') AND (NOT status = 'trash') AND (date_expire < NOW())", $this->table_name);
			$rows_affected = $wpdb->query($sql);

			return $rows_affected;
		}

		// Get submission count by status
		function db_get_count_by_status($form_id = 0, $status = '') {

			// User capability check
			if(!WS_Form_Common::can_user('read_submission')) { return false; }

			if(!WS_Form_Common::check_form_status($status, false)) { $status = ''; }
			if($form_id == 0) { return 0; }

			global $wpdb;

			$sql = sprintf("SELECT COUNT(id) FROM %s WHERE", $this->table_name);
			if($status == '') { $sql .= " NOT(status = 'trash')"; } else { $sql .= " status = '$status'"; }
			$sql .= " AND form_id = $form_id;";

			$form_count = $wpdb->get_var($sql);
			if(is_null($form_count)) { $form_count = 0; }

			return $form_count; 
		}

		// Get submit meta
		public function db_get_submit_meta($submit) {

			// No capabilities required, this is a public method

			$submit_meta = array();

			// Get submit record ID
			$submit_id = $submit['id'];

			// Read meta
			$ws_form_submit_meta = New WS_Form_Submit_Meta();
			$ws_form_submit_meta->parent_id = $submit_id;
			$meta_array = $ws_form_submit_meta->db_read_all();

			// Process meta data
			foreach($meta_array as $index => $meta) {

				// Get field value
				$value = (@unserialize($meta['meta_value']) !== false) ? unserialize($meta['meta_value']) : $meta['meta_value'];

				// Get field ID
				$field_id = intval($meta['field_id']);

				// If field ID found, process and return as array including type
				if($field_id > 0) {

					// Load field data to cache
					if(isset($field_cache[$field_id])) {

						// Use cached version
						$field = $field_cache[$field_id];

					} else {

						// Read field data and get type
						$ws_form_field = New WS_Form_Field();
						$ws_form_field->id = $field_id;
						$field = $ws_form_field->db_read(true);
						$field_cache[$field_id] = $field;
					}

					// Process according to type
					switch($field['type']) {

						case 'datetime' :

							$value_presentation_full = '';

							if($value != '') {

								$input_type_datetime = WS_Form_Common::get_object_meta_value((object)$field, 'input_type_datetime', 'date');

								$value_presentation_full = WS_Form_Common::get_date_by_type($value, $input_type_datetime);
							}

							$value = array('mysql' => $value, 'presentation_full' => $value_presentation_full);

							break;
					}

					// Return as array including type
					$submit_meta[WS_FORM_FIELD_PREFIX . $field_id] = array('id' => $field_id, 'value' => $value, 'type' => $field['type']);

				} else {

					// Return as string
					$submit_meta[$meta['meta_key']] = $value;
				}
			}

			return $submit_meta;
		}

		// Get number for form submissions
		public function db_get_count_submit_form() {

			// User capability check
			if(!WS_Form_Common::can_user('read_submission')) { return false; }

			// Check form ID
			self::db_check_form_id();

			global $wpdb;

			// Get total number for form submissions
			$sql = sprintf("SELECT COUNT(id) AS count_submit FROM %s WHERE form_id = %u;", $this->table_name, $this->form_id);
			$count_submit = $wpdb->get_var($sql);
			if(!is_null($count_submit)) { return absint($count_submit); } else { return 0; }
		}

		// Restore
		public function db_restore() {

			// User capability check
			if(!WS_Form_Common::can_user('delete_submission')) { return false; }

			self::db_set_status('draft');
		}

		// Set status of submit
		public function db_set_status($status) {

			// No capabilities required, this is a public method

			self::db_check_id();

			// Ensure provided form status is valid
			self::db_check_status($status);

			global $wpdb;

			// Update form record
			$sql = sprintf("UPDATE %s SET status = '%s' WHERE id = %u LIMIT 1;", $this->table_name, esc_sql($status), $this->id);
			if($wpdb->query($sql) === false) { parent::db_throw_error(__('Error setting submit status', 'ws-form')); }

			return true;
		}

		// Check form status
		public function db_check_status($status) {

			// Check status is valid
			$valid_statuses = explode(',', WS_FORM_STATUS_SUBMIT);
			if(!in_array($status, $valid_statuses)) { parent::db_throw_error(__('Invalid submit status: ' . $status, 'ws-form')); }

			return true;
		}

		// Get submit status name
		public function db_get_status_name($status) {

			switch($status) {

				case 'draft' : 		return __('In Progress', 'ws-form'); break;
				case 'publish' : 	return __('Submitted', 'ws-form'); break;
				case 'error' : 		return __('Error', 'ws-form'); break;
				case 'trash' : 		return __('Trash', 'ws-form'); break;
				default :			return $status;
			}
		}

		// Get submit columns
		public function db_get_submit_fields() {

			// User capability check
			if(!WS_Form_Common::can_user('read_submission')) { return false; }

			self::db_check_form_id();

			$visible_count = 0;
			$visible_count_max = 5;

			$submit_fields = array();

			// Create form object
			$ws_form_form = New WS_Form_Form();
			$ws_form_form->id = $this->form_id;

			// Read published form data
			$form_array = $ws_form_form->db_read_published();

			// Form not yet published
			if($form_array === false) { return array(); }

			// Convert to object
			$form_object = json_decode(json_encode($form_array));

			// Get fields in single dimension array
			$fields = WS_Form_Common::get_fields_from_form($form_object);

			// Get field types in single dimension array
			$field_types = WS_Form_Config::get_field_types_flat();

			// Excluded field types
			$field_types_excluded = array('textarea');

			foreach($fields as $field) {

				if(!isset($field_types[$field->type])) { continue; }

				// Get field type
				$field_type_config = $field_types[$field->type];

				// Skip unlicensed fields
				if(
					isset($field_type_config['pro_required']) &&
					$field_type_config['pro_required']

				) { continue; }

				// Skip fields that are not saved to meta data
				if(!$field_type_config['submit_save']) { continue; }

				// Determine if field is required
				$required = WS_Form_Common::get_object_meta_value($field, 'required', false);

				// Determine excluded fields
				$excluded = in_array($field->type, $field_types_excluded);

				// Push to submit_fields array
				$submit_fields[$field->id] = array(

					'label' 	=> $field->label,
					'required' 	=> $required,
					'excluded'	=> $excluded,
					'hidden'	=> true,
				);
			}

			// Go through each submit field and if it is required, mark it as not hidden
			foreach($submit_fields as $id => $field) {

				if($visible_count < $visible_count_max) {

					if($field['required'] && !$field['excluded']) {

						$submit_fields[$id]['hidden'] = false;
						$visible_count++;
					}

					if($visible_count == $visible_count_max) { break; }
				}
			}

			if($visible_count < $visible_count_max) {

				// Go through each submit field and if it is not required, mark it as not hidden
				foreach($submit_fields as $id => $field) {

					if($visible_count < $visible_count_max) {

						if(!$field['required'] && !$field['excluded']) {

							$submit_fields[$id]['hidden'] = false;
							$visible_count++;
						}

						if($visible_count == $visible_count_max) { break; }
					}
				}
			}

			return $submit_fields;
		}

		// Create CSV file
		public function db_export_csv(&$file, $submit_ids) {

			// User capability check
			if(!WS_Form_Common::can_user('export_submission')) { return false; }

			// Get field data
			$submit_fields = $this->db_get_submit_fields();

			// Build CSV column headings
			$csv_header = array();

			// Fixed fields
			$csv_header_fields = array('id' => 'Submission ID', 'status_full' => 'Status', 'date_added' => 'Date Added', 'date_updated' => 'Date Updated', 'user_id' => 'User ID', 'user_first_name' => 'User First Name', 'user_last_name' => 'User Last Name', 'duration' => 'Duration (Seconds)');
			foreach(array_slice($csv_header_fields, 1) as $key => $value) { $csv_header[] = $value; } // Ignore first index

			// Form fields
			foreach($submit_fields as $submit_field) { $csv_header[] = $submit_field['label']; }

			// Tracking
			$tracking = WS_Form_Config::get_tracking(false);
			foreach($tracking as $key => $tracking_config) {

				$csv_header[] = $tracking_config['label'];
			}

			// Output header row
			fwrite($file, '"ID",');	// To overcome issue with Excel thinking 'ID,' is an SYLK file
			fputcsv($file, $csv_header);

			// Get id's to download
			if(!is_array($submit_ids)) { $submit_ids = (empty($submit_ids) ? array() : array($submit_ids)); }

			if(count($submit_ids) > 0) {

				// Check integrity of array
				foreach($submit_ids as $key => $submit_id) {

					if(!is_numeric($submit_id)) { unset($submit_ids[$key]); }
				}

				// Throw error if no valid submit ID's
				if(count($submit_ids) == 0) { self::api_throw_error(__('Invalid submit ID')); }

				// Build WHERE sql
				$where = 'id IN (' . implode(',', $submit_ids) . ') AND ';

			} else {

				$where = '';
			}

			$where .= "NOT (status='trash')";

			$order_by = 'date_added';

			// Get data from API
			$submits = $this->db_read_all('', $where, $order_by);

			foreach($submits as $submit) {

				$row_array = array();

				// Fixed fields
				foreach($csv_header_fields as $key => $value) {

					switch($key) {

						case 'user_first_name' :

							$row_array[] = isset($submit['user']) ? $submit['user']['first_name'] : '';
							break;

						case 'user_last_name' :

							$row_array[] = isset($submit['user']) ? $submit['user']['last_name'] : '';
							break;

						default :

							$row_array[] = $submit[$key];
					}
				}

				// Form fields
				foreach($submit_fields as $id => $field) {

					$field_name = WS_FORM_FIELD_PREFIX . $id;

					// Get type
					$type = isset($submit[$field_name]) ? (isset($submit[$field_name]['type']) ? $submit[$field_name]['type'] : '') : '';

					// Get value
					$value = isset($submit[$field_name]) ? (isset($submit[$field_name]['value']) ? $submit[$field_name]['value'] : '') : '';

					// Apply filter
					$value = apply_filters('wsf_submit_field_type_csv', $value, $id, $type);

					// Process by type
					switch($type) {

						case 'signature' :
						case 'file' :

							if(!is_array($value)) { break; }

							$value_filenames = array();

							foreach($value as $value_file) {

								$value_filenames[] = $value_file['name'];
							}

							$value = implode(',', $value_filenames);

							break;

						case 'datetime' :

							if(is_array($value)) { $value = $value['mysql']; }
							break;
					}

					// Fix stray arrays
					if(is_array($value)) { $value = ''; }

					// Add column
					$row_array[] = $value;
				}

				// Tracking
				foreach($tracking as $key => $tracking_config) {

					// Get tracking value
					$value = isset($submit[$key]) ? $submit[$key] : '';

					// Add column
					$row_array[] = $value;
				}

				// Output data
				fputcsv($file, $row_array);
			}
		}

		// Create hash
		public function db_create_hash() {

			if($this->hash == '') { $this->hash = esc_sql(wp_hash($this->id . '_' . $this->form_id . '_' . time())); }

			return $this->hash;
		}

		// Check hash
		public function db_check_hash() {

			if(($this->hash == '') || (strlen($this->hash) != 32)) { self::db_throw_error(__('Invalid hash ID', 'ws-form')); }
			return true;
		}

		// Check form id
		public function db_check_form_id() {

			if($this->form_id <= 0) { self::db_throw_error(__('Invalid form ID', 'ws-form')); }
			return true;
		}

		// Check id
		public function db_check_id() {

			if($this->id <= 0) { self::db_throw_error(__('Invalid submit ID', 'ws-form')); }
			return true;
		}
	}