<?php

	class WS_Form_Wizard extends WS_Form_Core {

		public $id;
		public $label;
		public $form;
		public $wizard_category_id;
		public $pro_required;
		public $action_id;
		public $svg;

		private $table_name;
		private $table_name_meta;

		const DB_INSERT = 'id,label,form,wizard_category_id,pro_required,svg,sort_index';
		const DB_UPDATE = 'label,form';
		const DB_SELECT = 'id,label,form,wizard_category_id,pro_required,svg';

		public function __construct() {

			global $wpdb;

			$this->id = 0;
			$this->wizard_category_id = 0;
			$this->table_name = $wpdb->prefix . WS_FORM_DB_TABLE_PREFIX . 'wizard';
			$this->pro_required = false;
			$this->action_id = false;
			$this->svg = '';
		}

		// Create field
		public function db_create() {

			global $wpdb;

			// Get sort_index
			$sort_index = self::object_sort_index_get($this->table_name, 'wizard_category_id', $this->wizard_category_id);

			// Add wizard
			$sql = sprintf("INSERT INTO %s (%s) VALUES ('%s', '%s', %u, NOW(), NOW(), %u, %u, 
				'%s', %u);", $this->table_name, self::DB_INSERT, $this->id, esc_sql($this->label), esc_sql($this->form), $this->wizard_category_id, ($this->pro_required ? 1 : 0), $this->svg, $sort_index);
			if($wpdb->query($sql) === false) { self::db_throw_error(__('Error adding field', 'ws-form')); }

			// Get inserted ID
			$this->id = $wpdb->insert_id;
		}

		// Read record to array
		public function db_read() {

			global $wpdb;

			// Add fields
			$sql = sprintf("SELECT %s FROM %s WHERE id = %u LIMIT 1;", self::DB_SELECT, $this->table_name, $this->id);
			$wizard = $wpdb->get_row($sql, 'ARRAY_A');

			if($wizard === null) { self::db_throw_error(__('Unable to read field', 'ws-form')); }

			foreach($wizard as $key => $value) {

				$this->{$key} = $value;
			}

			// Decode form
			$this->form = json_decode($this->form);
			$this->form = (array)$this->form;

			// Return array
			return $wizard;
		}

		// Read all by db_check_wizard_category_id
		public function db_read_all() {

			global $wpdb;

			self::db_check_wizard_category_id();

			// Get form wizards
			$sql = sprintf("SELECT %s FROM %s WHERE wizard_category_id = %u ORDER BY sort_index;", self::DB_SELECT, $this->table_name, $this->wizard_category_id);
			$wizards = $wpdb->get_results($sql, 'ARRAY_A');

			return $wizards ? $wizards : [];
		}

		// Read all
		public function db_read_all_categories() {

			global $wpdb;

			$return_array = [];

			// Get form wizard categories
			$sql = sprintf("SELECT id, label FROM %s_category ORDER BY sort_index;", $this->table_name);
			$wizard_categories = $wpdb->get_results($sql, 'ARRAY_A');

			if($wizard_categories) {

				foreach($wizard_categories as $wizard_category) {

					// Set category ID
					$this->wizard_category_id = $wizard_category['id'];

					// Read wizards for this category
					$wizards = self::db_read_all();

					// Add to return array
					$return_array[] = array(

						'id'		=>	$this->wizard_category_id,
						'label'		=>	$wizard_category['label'],
						'wizards'	=>	$wizards,
					);
				}
			}

			return $return_array;
		}

		// Get wizards for each action installed
		public function db_get_actions() {

			$return_array = array();

			if(!isset(WS_Form_Action::$actions)) { parent::db_throw_error(__('No actions installed', 'ws-form')); }

			// Capabilities required of each action
			$capabilities_required = array('get_lists', 'get_list', 'get_list_fields');

			// Get actions that have above capabilities
			$actions = WS_Form_Action::get_actions_with_capabilities($capabilities_required);

			// Run through each action
			foreach($actions as $action) {

				// Add to return array
				$return_array[] = array(

					'id'		=>	$action->id,
					'label'		=>	$action->label,
					'reload'	=>	isset($action->add_new_reload) ? $action->add_new_reload : true
				);
			}

			return $return_array;
		}

		// Get wizards for each action installed
		public function db_get_action_wizards() {

			$return_array = array();

			if(!isset(WS_Form_Action::$actions)) { parent::db_throw_error(__('No actions installed', 'ws-form')); }

			// Check action ID
			self::db_check_action_id();

			// Capabilities required of each action
			$capabilities_required = array('get_lists', 'get_list', 'get_list_fields');

			// Get actions that have above capabilities
			$actions = WS_Form_Action::get_actions_with_capabilities($capabilities_required);

			if(!isset($actions[$this->action_id])) { parent::db_throw_error(__('Action not compatible with this function', 'ws-form')); }

			$action = $actions[$this->action_id];

			// Get lists
			$lists = $action->get_lists();

			foreach($lists as $list) {

				// Add to return array
				$return_array[] = array(

					'id'			=>	$list['id'],
					'label'			=>	$list['label'],
					'field_count'	=>	$list['field_count'],
					'record_count'	=>	$list['record_count'],
					'svg'			=>	WS_Form_Action::get_svg($this->action_id, $list['id'], $list['label'], $list['field_count'], $list['record_count'])
				);
			}

			return $return_array;
		}

		// Check wizard_category_id
		public function db_check_wizard_category_id() {

			if($this->wizard_category_id <= 0) { parent::db_throw_error(__('Invalid wizard category ID', 'ws-form')); }
			return true;
		}

		// Check action_id
		public function db_check_action_id() {

			if($this->action_id === false) { parent::db_throw_error(__('Invalid action ID', 'ws-form')); }
			return true;
		}
	}