<?php

	/**
	 * Form preview
	 */
	class WS_Form_Preview {

		protected $form_id = 0;
		protected $ws_form_form;

		public function __construct() {

			// Get form_id
			$this->form_id = isset($_GET) ? isset($_GET['wsf_preview_form_id']) ? intval($_GET['wsf_preview_form_id']) : 0 : 0;
			if($this->form_id <= 0) { return false; }

			// Load form
			$this->ws_form_form = New WS_Form_Form();
			$this->ws_form_form->id = $this->form_id;
			$this->ws_form_form->db_read(false);

			// Actions
			add_action('pre_get_posts', array($this, 'pre_get_posts'));

			// Page title
			add_filter('the_title', array($this, 'the_title'));

			// Content
			remove_filter('the_content', 'wpautop');
			add_filter('the_content', array($this, 'the_content'), 1);

			remove_filter('the_excerpt', 'wpautop');
			add_filter('get_the_excerpt', array($this, 'the_content'));

			add_filter('template_include', array($this, 'template_include'));
			add_filter('post_thumbnail_html', array($this, 'post_thumbnail_html'));
		}

		public function pre_get_posts($query) {

			$query->set('posts_per_page', 1);
		}

		function the_title($title) {

			// Check user is logged in
			if(!is_user_logged_in()) { return false; }

			if(!in_the_loop()) return $title;

			return $this->ws_form_form->label . ' ' . __('Preview', 'ws-form');
		}

		function the_content() {

			return do_shortcode('[' . WS_FORM_SHORTCODE . ' id=' . $this->form_id . ' published="false" preview="true"]');
		}

		function template_include() {

			$templates_path = get_template_directory();

			// Get preview template
			$preview_template = WS_Form_Common::option_get('preview_template', '');

			// Template selected in settings, so use that
			if($preview_template != '') { return $templates_path . '/' . $preview_template; }

			// Load templates until we find one that contains the_content()
			$templates = array('page.php', 'singular.php', 'index.php');

			// Return a template that contains the_content()
			foreach($templates as $template) {

				// Build template path
				$template_file = $templates_path . '/' . $template;

				// Skip files that don't exist
				if(!file_exists($template_file)) { continue; }

				// Load template into string
				$template_html = file_get_contents($template_file);

				// Look for the_content(
				if(strpos($template_html, 'the_content(') !== false) {

					return $template_file;
				}
			}

			// Fallback
			return WS_FORM_PLUGIN_DIR_PATH . 'public/preview-fallback.php';
		}

		function post_thumbnail_html() {

			return '';
		}
	}
