<?php

	/**
	 * The public-facing functionality of the plugin.
	 *
	 * Defines the plugin name, version, and two examples hooks for how to
	 * enqueue the admin-specific stylesheet and JavaScript.
	 */

	class WS_Form_Public {

		// The ID of this plugin.
		private $plugin_name;

		// The version of this plugin.
		private $version;

		// Customizer
		private $customizer;

		// Form index (Incremented for each rendered with a short code)
		public $form_instance = 1;

		// Error index
		public $error_index = 0;

		// Debug
		public $debug = false;

		// JSON arrays
		public $wsf_form_json = array();

		// Footer JS
		public $footer_js = '';

		// Initialize the class and set its properties.
		public function __construct($plugin_name, $version) {

			$this->plugin_name = $plugin_name;
			$this->version = $version;
			$this->customizer = (WS_Form_Common::get_query_var('customize_theme') !== '');
		}

		// Register the stylesheets for the public-facing side of the site.
		public function enqueue_styles() {


			// CSS - Layout
			if((WS_Form_Common::option_get('css_layout', true)) && (WS_Form_Common::option_get('framework', 'ws-form') == 'ws-form')) {

				wp_enqueue_style($this->plugin_name . '-css-layout', WS_Form_Common::get_api_base_path() . 'helper/ws_form_css', array(), $this->version, 'all');
			}


			// CSS - Skin
			if((WS_Form_Common::option_get('css_skin', true)) && (WS_Form_Common::option_get('framework', 'ws-form') == 'ws-form')) {

				if($this->customizer) {

					add_action('wp_head', array($this, 'public_css_inline'));

				} else {

					wp_enqueue_style($this->plugin_name . '-css-skin', WS_Form_Common::get_api_base_path() . 'helper/ws_form_css_skin', array(), $this->version, 'all');
				}
			}
		}

		// Public CSS - Render inline (For customizer only)
		public function public_css_inline() {

			echo '<style>';

			// Output CSS
			$ws_form_css = new WS_Form_CSS();
			$ws_form_css->render_skin();

			echo '</style>';
		}

		// Register the JavaScript for the public-facing side of the site.
		public function enqueue_scripts() {

			// Enqueued scripts settings
			$ws_form_settings = self::localization_object($this->debug);

			// Form class
			wp_register_script($this->plugin_name . '-form-common', plugin_dir_url(__DIR__) . 'shared/js/ws-form.js', array('jquery'), $this->version, false);
			wp_localize_script($this->plugin_name . '-form-common', 'ws_form_settings', $ws_form_settings);
			wp_enqueue_script($this->plugin_name . '-form-common');

			// Form class - Public
			wp_register_script($this->plugin_name . '-form-public', plugin_dir_url(__DIR__) . 'public/js/ws-form-public.js', array('jquery'), $this->version, false);
			wp_enqueue_script($this->plugin_name . '-form-public');

		}

		/* Shortcode: ws_form */
		public function shortcode_ws_form($atts) {

			// Read form ID
			if(isset($atts['id'])) { $form_id = absint($atts['id']); } else { return ''; }
			$published = isset($atts['published']) ? ($atts['published'] == 'true') : true;

			// Check for preview mode
			if(isset($atts['preview'])) {

				// Reset form instance (This is required to ensure wp_head calls resulting in do_shortocde('ws-form') don't stack up on each other)
				$this->form_instance = 1;
			}

			if($form_id > 0) {

				// Embed form data (Avoids an API call)
				$ws_form_form = New WS_Form_Form();
				$ws_form_form->id = $form_id;

				try {

					if($published) {

						$form_array = $ws_form_form->db_read_published();

					} else {

						$form_array = $ws_form_form->db_read(true, true);
					}

				} catch(Exception $e) { return $e->getMessage(); }

				// Get form HTML
				$return_value = self::form_html($this->form_instance++, $form_id, $form_array, $published);

				return $return_value;

			} else {

				// Error
				return __('Invalid form ID', 'ws-form');
			}
		}

		// Footer scripts
		public function wp_footer() {

			echo "<script>\n\n";

			// Embed config data (Avoids an API call)
			$json_config = json_encode(WS_Form_Config::get_config());
			echo sprintf("\tvar wsf_form_json_config = %s;\n", $json_config);

			// Init form data
			echo "var wsf_form_json = [];\n";
			echo "var wsf_form_json_populate = [];\n\n";

			// Footer JS
			echo $this->footer_js;

			echo "</script>\n\n";
		}

		// Form - HTML
		public function form_html($form_instance, $form_id, $form_array, $published = true) {

			if($form_array === false) { return __('Unpublished form', 'ws-form'); }
			if(!is_array($form_array)) { return __('Invalid form data', 'ws-form'); }

			// Do not render if draft, future or trash
			switch($form_array['status']) {

				case 'draft' :
				case 'future' :
				case 'trash' :

					if($published) { return ''; };
			}

			// Check for form attributes
			$form_attributes = '';

			$framework_id = WS_Form_Common::option_get('framework', WS_FORM_DEFAULT_FRAMEWORK);

			$frameworks = WS_Form_Config::get_frameworks();

			$framework = $frameworks['types'][$framework_id];

			if(isset($framework['form']['public']['attributes'])) {

				$form_attribute_array = $framework['form']['public']['attributes'];

				foreach($form_attribute_array as $key => $value) {

					$form_attributes .= ' ' . $key . (($value != '') ? '="' . $value . '"' : '');
				}
			}

			// Preview attribute
			if(!$published) { $form_attributes .= ' data-preview'; }

			// CSS - Framework
			if((isset($framework['css_file'])) && ($framework['css_file'] != '')) {

				$css_file_path = plugin_dir_url(__FILE__) . 'css/framework/' . $framework['css_file'];
				wp_enqueue_style($this->plugin_name . '-css-framework', $css_file_path, array(), $this->version, 'all');
			}

			// Form wrapper
			$return_value = sprintf("<form class=\"wsf-form\" id=\"ws-form-%1\$u\" data-id=\"%2\$u\" data-instance-id=\"%1\$u\" method=\"post\" novalidate%3\$s></form>", $form_instance, $form_id, $form_attributes);

			// Build JSON
			$form_json = json_encode($form_array);

			// Form data (Only render once per form ID)
			if(!isset($wsf_form_json[$form_id])) {

				// Form JSON
				$this->footer_js .= sprintf("wsf_form_json[%u] = %s;", $form_id, $form_json);

				// Form JSON populate
				$populate_array = self::get_populate_array($form_json);
				if(($populate_array !== false) && count($populate_array) > 0) {

					$this->footer_js .= sprintf("wsf_form_json_populate[%u] = %s;", $form_id, json_encode($populate_array));
				}

				$wsf_form_json[$form_id] = true;
			}

			return $return_value;
		}

		public function localization_object($debug = false) {

			global $post;

			$return_array = array(

				// Nonce
				'nonce'					=>	wp_create_nonce('wp_rest'),

				// URL
				'url'					=>	WS_Form_Common::get_api_base_path(),

				// Admin URL
				'admin_url'				=>	admin_url(),

				// Includes URL
				'includes_url'			=>	includes_url(),

				// Admin framework (Fallover)
				'framework_admin'		=> 'ws-form',

				// Should framework CSS be rendered? (WS Form framework only)
				'css_framework'			=> WS_Form_Common::option_get('css_framework', true),


				// Is design enabled?
				'design'				=> (WS_Form_Common::option_get('css_skin', true)) && (WS_Form_Common::option_get('framework', 'ws-form') == 'ws-form'),

				// Max upload size
				'max_upload_size'		=> absint(WS_Form_Common::option_get('max_upload_size', 0)),

				// Field prefix
				'field_prefix'			=> WS_FORM_FIELD_PREFIX
			);

			// Pass through post ID
			if(isset($post) && ($post->ID > 0)) {

				$return_array['post_id'] = $post->ID;
			}

			return $return_array;
		}

		// Populate from action
		public function get_populate_array($form_json) {

			// Get populate data
			$form_object = json_decode($form_json);

			// Check form populate is enabled
			$form_populate_enabled = WS_Form_Common::get_object_meta_value($form_object, 'form_populate_enabled', '');
			if(!$form_populate_enabled) { return false; }

			// Read form populate data - Action ID
			$form_populate_action_id = WS_Form_Common::get_object_meta_value($form_object, 'form_populate_action_id', '');
			if($form_populate_action_id == '') { return false; }
			if(!isset(WS_Form_Action::$actions[$form_populate_action_id])) { return false; }

			// Get action
			$action = WS_Form_Action::$actions[$form_populate_action_id];

			// Check get method exists
			if(!method_exists($action, 'get')) { return false; }

			// Read form populate data - List ID
			$form_populate_list_id = WS_Form_Common::get_object_meta_value($form_object, 'form_populate_list_id', '');
			if($form_populate_list_id == '') { return false; }

			// Read form populate data - Field mapping
			$form_populate_field_mapping = WS_Form_Common::get_object_meta_value($form_object, 'form_populate_field_mapping', array());

			if(method_exists($action, 'get_tags')) {

				// Read form populate data - Tag mapping
				$form_populate_tag_mapping = WS_Form_Common::get_object_meta_value($form_object, 'form_populate_tag_mapping', array());
			}

			// Get user data
			$current_user = wp_get_current_user();

			// Is user logged in?
			if($current_user->ID === 0) { return false; }

			// Get user email address
			$user_email = $current_user->user_email;

			// Set list ID
			$action->list_id = $form_populate_list_id;

			// Try to get user data
			try {

				$get_array = $action->get($user_email);

			} catch(Exception $e) { return false; }

			$return_array = array();

			// Process field mapping data
			$field_mapping_lookup = array();
			foreach($form_populate_field_mapping as $field_mapping) {

				$action_field = $field_mapping->form_populate_list_fields;
				$ws_form_field = $field_mapping->ws_form_field;

				if(!isset($field_mapping_lookup[$action_field])) {

					$field_mapping_lookup[$action_field] = $ws_form_field;
				}
			}

			// Map fields
			if(isset($get_array['fields'])) {

				foreach($get_array['fields'] as $id => $value) {

					if(isset($field_mapping_lookup[$id])) {

						$return_array[WS_FORM_FIELD_PREFIX . $field_mapping_lookup[$id]] = $value;
					}
				}
			}

			if(method_exists($action, 'get_tags')) {

				// Process tag mapping data
				$tag_mapping_array = array();
				foreach($get_array['tags'] as $id => $value) {

					if($value) {

						$tag_mapping_array[] = $id;
					}
				}

				// Map tags
				foreach($form_populate_tag_mapping as $tag_mapping) {

					$ws_form_field = $tag_mapping->ws_form_field;

					$return_array[WS_FORM_FIELD_PREFIX . $ws_form_field] = $tag_mapping_array;
				}
			}

			return array('action_label' => $action->label, 'data' => $return_array);
		}
	}
