(function($) {

	'use strict';

	// Set is_admin
	$.WS_Form.prototype.set_is_admin = function() { return false; }

	// One time init for admin page
	$.WS_Form.prototype.init = function() {

		// Build data cache
		this.data_cache_build();

		// Set global variables once for performance
		this.set_globals();
	}

	// Continue initialization after submit data retrieved
	$.WS_Form.prototype.init_after_get_submit = function(submit_retrieved) {


		// Build form
		this.form_build();
	}

	// Set global variables once for performance
	$.WS_Form.prototype.set_globals = function() {

		// Get framework ID
		this.framework_id = $.WS_Form.settings_plugin.framework;

		// Get framework settings
		this.framework = $.WS_Form.frameworks.types[this.framework_id];

		// Get current framework
		this.framework_fields = this.framework['fields']['public'];

		// Get invalid_feedback placeholder mask
		this.invalid_feedback_mask_placeholder = '';
		if(typeof($.WS_Form.meta_keys['invalid_feedback']) !== 'undefined') {

			if(typeof $.WS_Form.meta_keys['invalid_feedback']['mask_placeholder'] !== 'undefined') {
				
				this.invalid_feedback_mask_placeholder = $.WS_Form.meta_keys['invalid_feedback']['mask_placeholder'];
			}
		}

		// Get activated class
		var class_validated_array = (typeof this.framework.fields.public.class_form_validated !== 'undefined') ? this.framework.fields.public.class_form_validated : [];
		this.class_validated = class_validated_array.join(' ');


		// Set hash (Reset cookie with new expiry)
		this.hash_set(this.cookie_get('hash', ''), true);

		// Read submission data if hash is defined
		var ws_this = this;
		if(this.hash != '') {

			// Call AJAX request
			$.WS_Form.this.api_call('submit/hash/' + this.hash, 'GET', {}, function(response) {

				// Save the submissions data
				ws_this.submit = response.data;

				ws_this.init_after_get_submit(true);

			}, function(response) {

				// Read auto populate data instead
				ws_this.read_json_populate();

				ws_this.init_after_get_submit(false);
			});

		} else {

			// Read auto populate data
			this.read_json_populate();

			this.init_after_get_submit(false);
		}
	}

	// Read auto populate data
	$.WS_Form.prototype.read_json_populate = function() {

		if(typeof(wsf_form_json_populate) !== 'undefined') {

			if(typeof(wsf_form_json_populate[this.form_id]) !== 'undefined') {

				this.submit_auto_populate = wsf_form_json_populate[this.form_id]['data'];
			}
		}
	}


	// Render any interface elements that rely on the form object
	$.WS_Form.prototype.form_render = function() {

		// Initialize JS
		this.form_init_js();

		// Groups - Tabs - Initialize
		this.form_tabs();


		// Navigation
		this.form_navigation();

		// Client side form validation
		this.form_validation();


		// Required
		this.form_required();


		// Text input and textarea character and word count
		this.form_character_word_count();

		// Spam protection
		this.form_spam_protection();


		// Form preview
		this.form_preview();

		// Trigger rendered event
		this.trigger('rendered');

	}

	// Trigger events
	$.WS_Form.prototype.trigger = function(slug) {

		// Instance event
		var trigger_instance = 'wsf-' + slug + '-instance-' + this.form_instance_id;
		$(window).trigger(trigger_instance);

		// Form event
		var trigger_form = 'wsf-' + slug + '-form-' + this.form_id;
		$(window).trigger(trigger_form);
	}

	// Initialize JS
	$.WS_Form.prototype.form_init_js = function() {

		// Check framework init_js
		if(typeof this.framework.init_js !== 'undefined') {

			// Framework init JS values
			var framework_init_js_values = {'form_selector': '#' + this.form_obj_id};
			var framework_init_js = this.mask_parse(this.framework.init_js, framework_init_js_values);

			$.globalEval('(function($) { $(function() {' + framework_init_js + '}) })(jQuery);');
		}
	}

	// Form - Reset
	$.WS_Form.prototype.form_reset = function() {

		// HTML form reset
		this.form_obj[0].reset();

	}

	// Form - Clear
	$.WS_Form.prototype.form_clear = function() {

		// Clear fields
		for(var key in this.field_data_cache) {

			var field = this.field_data_cache[key];

			// Only process on submit save fields
			var submit_save = this.get_field_value_fallback(field.type, false, 'submit_save', false);
			if(!submit_save) { continue; }

			var field_id = field.id;
			var field_selector = '#' + this.form_id_prefix + 'field-' + field.id;
			var field_name = this.field_name_prefix + field_id;

			// Clear value
			switch(field.type) {

				case 'checkbox' :
				case 'radio' :

					$('[name="' + field_name + '[]"]').prop('checked', false);
					break;

				case 'select' :

					$('[name="' + field_name + '[]"]').val('');
					break;

				default:

					$(field_selector).val('');
			}

			$(field_selector).trigger('change').trigger('input');
		}

	}

	// Form reload
	$.WS_Form.prototype.form_reload = function() {

		// Read submission data if hash is defined
		var ws_this = this;
		if(this.hash != '') {

			// Call AJAX request
			$.WS_Form.this.api_call('submit/hash/' + this.hash, 'GET', {}, function(response) {

				// Save the submissions data
				ws_this.submit = response.data;

				ws_this.form_reload_after_get_submit(true);

			}, function(response) {

				ws_this.form_reload_after_get_submit(false);
			});

		} else {

			// Reset submit
			this.submit = false;

			this.form_reload_after_get_submit(false);
		}
	}

	// Form reload - After get submit
	$.WS_Form.prototype.form_reload_after_get_submit = function(submit_retrieved) {

		// Reset form tag
		this.form_obj.removeClass(this.class_validated)

		// Empty form object
		this.form_obj.empty();

		// Build form
		this.form_build();
	}

	// Form - Hash reset
	$.WS_Form.prototype.form_hash_clear = function() {

		// Clear hash variable
		this.hash = '';

		// Clear hash cookie
		this.cookie_clear('hash')

	}


	// Form navigation
	$.WS_Form.prototype.form_navigation = function() {

		var ws_this = this;
		var group_count = this.form.groups.length;

		// Buttons - Save
		$('[data-action="wsf-save"]', this.form_obj).each(function() {

			// Click
			$(this).click(function() {

				ws_this.form_post(false);
			});
		});


		// Buttons - Next
		$('[data-action="wsf-tab_next"]', this.form_obj).each(function() {

			// Get wrapper
			var button_wrapper = $(this).closest('[data-type="tab_next"]');

			// If there are no groups, remove the next button
			if(group_count == 0) { button_wrapper.remove(); }

			// Get group index
			var group_index = ws_this.get_group_index($(this));
			if(group_index === false) { return true; }

			// If the next button is on the last tab, disable it
			if(group_index == (group_count - 1)) { $(this).attr('disabled', true); return true; }

			// Click
			$(this).click(function() {

				ws_this.group_index_offset($(this), 1);
			});
		});

		// Buttons - Previous
		$('[data-action="wsf-tab_previous"]', this.form_obj).each(function() {

			// Get wrapper
			var button_wrapper = $(this).closest('[data-type="tab_previous"]');

			// If there are no groups, remove the previous button
			if(group_count == 0) { button_wrapper.remove(); }

			// Get group index
			var group_index = ws_this.get_group_index($(this));
			if(group_index === false) { return true; }

			// If the previous button is on the first tab, disable it
			if(group_index == 0) { $(this).attr('disabled', true); return true; }

			// Click
			$(this).click(function() {

				ws_this.group_index_offset($(this), -1);
			});
		});

		// Buttons - Reset
		$('[data-action="wsf-reset"]', this.form_obj).each(function() {

			// Click
			$(this).click(function() {

				ws_this.form_reset();
			});
		});

		// Buttons - Clear
		$('[data-action="wsf-clear"]', this.form_obj).each(function() {

			// Click
			$(this).click(function() {

				ws_this.form_clear();
			});
		});
	}

	// Tab - Activate by offset amount
	$.WS_Form.prototype.group_index_offset = function(obj, group_index_offset) {

		// Get group index
		var group_index = this.get_group_index(obj);

		// Get next group_index
		var group_index_new = group_index + group_index_offset;

		// Activate tab
		this.group_index_set(group_index_new);
	}

	// Tab - Set
	$.WS_Form.prototype.group_index_set = function(group_index) {

		if(this.form.groups.length <= 1) { return false; }

		var framework_tabs = this.framework['tabs']['public'];

		if(typeof framework_tabs.activate_js !== 'undefined') {

			var activate_js = framework_tabs.activate_js;	

			if(activate_js != '') {

				// Parse activate_js
				var mask_values = {'form': '#' + this.form_obj_id, 'index': group_index};
				var activate_js_parsed = this.mask_parse(activate_js, mask_values);

				// Execute activate tab javascript
				$.globalEval('(function($) { $(function() {' + activate_js_parsed + '}); })(jQuery);');

				// Set cookie
				this.cookie_set('tab_index', group_index);
			}
		}

	}

	// Get tab index object resides in
	$.WS_Form.prototype.get_group_index = function(obj) {

		if(this.form.groups.length <= 1) { return false; }

		// Get group
		var group_single = obj.closest('[data-group-index]');
		if(group_single.length == 0) { return false; }

		// Get group index
		var group_index = group_single.first().attr('data-group-index');
		if(group_index == undefined) { return false; }

		return parseInt(group_index);
	}


	// Form preview
	$.WS_Form.prototype.form_preview = function() {

		if(this.form_obj[0].hasAttribute("data-preview")) {

			this.form_obj.append('<input type="hidden" name="preview" value="true" />');
		}
	}

	// Form spam protection
	$.WS_Form.prototype.form_spam_protection = function() {

		// Honeypot
		var honeypot = this.get_object_meta_value(this.form, 'honeypot', false);

		if(honeypot) {

			// Add honeypot field
			var honeypot_hash = (this.form.published_checksum != '') ? this.form.published_checksum : 'honeypot_unpublished';

			// Build honeypot input
			var honeypot_input = '<input type="text" name="field_' + honeypot_hash + '" value="" />';

			// Add to form
			$(this.form_obj).append(honeypot_input);

			// Hide it
			var honeypot_obj = $('[name="field_' + honeypot_hash + '"]', this.form_obj);
			honeypot_obj.css({'position': 'absolute', 'left': '-9999em'});

		}
	}


	// Form analytics - Process - Fields
	$.WS_Form.prototype.form_analytics_process_fields = function(type, type_function) {

		var ws_this = this;

		// Run through all fields and set up analytics events for this type and function
		var analytics_config = $.WS_Form.analytics[type].functions[type_function];
		var analytics_label = $.WS_Form.analytics[type].label;
		var analytics_function_label = $.WS_Form.analytics[type].functions[type_function].label;

		for(var field_index in this.field_data_cache) {

			var field_type = this.field_data_cache[field_index].type;

			if(typeof $.WS_Form.field_type_cache[field_type].analytics === 'undefined') { continue; }

			var analytics_event = $.WS_Form.field_type_cache[field_type].analytics.event;
			var field_id = this.field_data_cache[field_index].id;
			var field_obj = $('#' + this.form_id_prefix + 'field-' + field_id);

			// Create event
			field_obj.on(analytics_event, function() {

				if($(this).attr('data-analytics-event-fired') === undefined) {

					var field_wrapper_obj = $(this).closest('[data-type]');
					var field_id = field_wrapper_obj.attr('data-id');
					var field = ws_this.field_data_cache[field_id];
					var analytics_event_category = $.WS_Form.field_type_cache[field.type].analytics.event_category;

					// Parse fields
					var analytics_event_values = {

						'form_label': ws_this.js_string_encode(ws_this.form.label),
						'event_category': analytics_event_category,
						'object_label': ws_this.js_string_encode(field.label)
					}

					// Parse event field args
					var analytics_event_function = atob(analytics_config.analytics_event_function);
					var analytics_event_function_parsed = ws_this.mask_parse(analytics_event_function, analytics_event_values);

					// Call function
					$.globalEval('(function($) {' + analytics_event_function_parsed + '})(jQuery);');

					// Log event
					ws_this.log('log_analytics_event_field_fired', analytics_label + ' (' + analytics_function_label + ') ' + analytics_event_function_parsed, 'analytics');

					$(this).attr('data-analytics-event-fired', 'true');
				}
			});
		}

		// Log event
		ws_this.log('log_analytics_event_field', analytics_label + ' (' + analytics_function_label + ')', 'analytics');
	}

	// Adds required string (if found in framework config) to all labels
	$.WS_Form.prototype.form_required = function() {

		// Get required label HTML
		var label_required = this.get_object_meta_value(this.form, 'label_required', false);
		if(!label_required) { return false; }

		var label_mask_required = this.get_object_meta_value(this.form, 'label_mask_required', '', true, true);
		if(label_mask_required == '') {

			// Use framework mask_required_label
			var framework_type = $.WS_Form.settings_plugin.framework;
			var framework = $.WS_Form.frameworks.types[framework_type];
			var fields = this.framework['fields']['public'];

			if(typeof fields.mask_required_label === 'undefined') { return false; }
			var label_mask_required = fields.mask_required_label;
			if(label_mask_required == '') { return false; }
		}

		// Get all labels in this form
		$('label', this.form_obj).each(function() {

			// Get 'for' attribute of label
			var label_for = $(this).attr('for');
			if(label_for === undefined) { return; }

			// Get field related to 'for'
			var field = $('[id="' + label_for + '"]');
			if(!field.length) { return; }

			// Check if field is required
			var field_required = field.attr('required');

			// Check if the require string should be added to the parent label (e.g. for radios)
			var label_required_id = $(this).attr('data-label-required-id');
			if((typeof(label_required_id) !== 'undefined') && (label_required_id !== false)) {

				var label_obj = $('#' + label_required_id);

			} else {

				var label_obj = $(this);
			}

			// Check if wsf-required-wrapper span exists, if not, create it (You can manually insert it in config using #required)
			var required_wrapper = $('.wsf-required-wrapper', label_obj);
			if(!required_wrapper.length && field_required) {

				var required_wrapper_html = '<span class="wsf-required-wrapper"></span>';

				// Find first child
				var first_child = label_obj.children().first();

				// Add at appropriate place
				if(first_child.length) {

					first_child.before(required_wrapper_html);

				} else {

					label_obj.append(required_wrapper_html);
				}

				required_wrapper = $('.wsf-required-wrapper', label_obj);
			}

			if(field_required) {

				// Add it
				required_wrapper.html(label_mask_required);

			} else {

				// Remove it
				required_wrapper.html('');
			}
		});
	}

	// Form - Client side validation
	$.WS_Form.prototype.form_validation = function() {

		// WS Form forms are set with novalidate attribute so we can manage that ourselves
		var ws_this = this;

		// Remove any existing on events
		this.form_obj.off();

		// On submit
		this.form_obj.on('submit', function(event) {

			// Do not submit form
			event.preventDefault();
			event.stopPropagation();

			// If form post is locked, return
			if(ws_this.form_post_locked) { return; }

			// Mark as validated
			$(this).addClass(ws_this.class_validated);

			// Check validity of form
			if(ws_this.form_validate($(this))) {

					// Submit form
					ws_this.form_post(true);
			}
		});

		// On reset
		this.form_obj.on('reset', function(event) {

			// Unmark as validated
			$(this).removeClass(ws_this.class_validated);

		});
	}

	// Form - Validate (WS Form validation functions)
	$.WS_Form.prototype.form_validate = function(form) {

		// Trigger rendered event
		this.trigger('validate-before');

		// Tab focussing
		var group_index_focus_enabled = (this.form.groups.length > 0);
		var group_index_focus = false;
		var object_focus = false;

		// Get form as element
		var form_el = form[0];

		// Execute browser validation
		var form_validated = form_el.checkValidity();

		if(!form_validated) {

			// Get all invalid fields
			var fields_invalid = $(':invalid', form).not('fieldset');

			if(fields_invalid) {

				// Get first invalid field
				object_focus = fields_invalid.first();

				// Get group index
				group_index_focus = this.get_group_index(object_focus);
			}
		}

		// Focus
		if(!form_validated) {

			// Focus tab
			if(group_index_focus !== false) { this.group_index_set(group_index_focus); }

			if(object_focus !== false) {

				// Focus object
	 			if(this.get_object_meta_value(this.form, 'invalid_field_focus', true)) {

					object_focus.focus();
				}

				// Scroll object
	 			if(this.get_object_meta_value(this.form, 'invalid_field_scroll', true)) {

					object_focus[0].scrollIntoView();
				}
			}
		}

		// Trigger rendered event
		this.trigger('validate-after');

		return form_validated;
	}


	// Convert hex color to RGB values
	$.WS_Form.prototype.conditional_hex_to_hsl = function(color) {

		// Get RGB of hex color
		var rgb = this.conditional_hex_to_rgb(color);
		if(rgb === false) { return false; }

		// Get HSL of RGB
		var hsl = this.conditional_rgb_to_hsl(rgb);

		return hsl;
	}

	// Convert hex color to RGB values
	$.WS_Form.prototype.conditional_hex_to_rgb = function(color) {

		// If empty, return false
		if(color == '') { return false; }

		// Does color have a hash?
		var color_has_hash = (color[0] == '#');

		// Check
		if(color_has_hash && (color.length != 7)) { return false; }
		if(!color_has_hash && (color.length != 6)) { return false; }

		// Strip hash
		var color = color_has_hash ? color.substr(1) : color;

		// Get RGB values
		var r = parseInt(color.substr(0,2), 16);
		var g = parseInt(color.substr(2,2), 16);
		var b = parseInt(color.substr(4,2), 16);

		return {'r': r, 'g': g, 'b': b};
	}

	// Convert RGB to HSL
	$.WS_Form.prototype.conditional_rgb_to_hsl = function(rgb) {

		if(typeof rgb.r === 'undefined') { return false; }
		if(typeof rgb.g === 'undefined') { return false; }
		if(typeof rgb.b === 'undefined') { return false; }

		var r = rgb.r;
		var g = rgb.g;
		var b = rgb.b;

		r /= 255, g /= 255, b /= 255;

		var max = Math.max(r, g, b), min = Math.min(r, g, b);
		var h, s, l = (max + min) / 2;

		if(max == min){
	
			h = s = 0;
	
		} else {
	
			var d = max - min;
			s = l > 0.5 ? d / (2 - max - min) : d / (max + min);

			switch(max){
				case r: h = (g - b) / d + (g < b ? 6 : 0); break;
				case g: h = (b - r) / d + 2; break;
				case b: h = (r - g) / d + 4; break;
			}

			h /= 6;
		}

		return {'h': h, 's': s, 'l': l};
	}

	$.WS_Form.prototype.conditional_process_actions = function(actions, type) {

		var actions_processed = 0;
		var process_required = false;

		for(var action_index in actions) {

			var action_single = actions[action_index];

			// Check integrity of action
			if(!this.conditional_action_check(action_single)) { continue; }

			// Read action data
			var object = action_single['object'];
			var object_id = action_single['object_id'];
			var object_row_id = (typeof action_single['object_row_id'] === 'undefined') ? false : action_single['object_row_id'];
			if(object_row_id < 1) { object_row_id = false; }
			var action = action_single['action'];
			var value = (typeof action_single['value'] === 'undefined') ? false : this.parse_variables_process(action_single['value']);
			var debug_action_value = value;
			var field_name = ws_form_settings.field_prefix + object_id;

			// Build selectors

			// Object wrapper
			switch(object) {

				case 'field' :

					var object_selector_wrapper = '#' + this.form_id_prefix + object + '-wrapper-' + object_id;
					break;

				default :

					var object_selector_wrapper = '#' + this.form_id_prefix + object + '-' + object_id;

			}
			var obj_wrapper = $(object_selector_wrapper);

			// Object
			var object_selector = '#' + this.form_id_prefix + object + '-' + object_id;
			var obj = $(object_selector);

			// Row
			if(object_row_id !== 0) {

				var object_selector_row = '#' + this.form_id_prefix + object + '-' + object_id + '-' + object_row_id;
				var obj_row = $(object_selector_row);
			}

			switch(action) {

				// Set value
				case 'value' :
				case 'value_number' :
				case 'value_datetime' :
				case 'value_color' :
				case 'value_tel' :
				case 'value_number' :
				case 'value_email' :
				case 'value_textarea' :

					obj.val(value).trigger('change').trigger('input').trigger('keyup');
					break;

				// Set HTML
				case 'html' :

					obj_wrapper.html(value);
					break;

				// Set button label
				case 'button_html' :

					obj.html(value);
					break;

				// Add class (Wrapper)
				case 'class_add_wrapper' :

					obj_wrapper.addClass(value);
					debug_action_value = this.language('debug_action_added');
					break;

				// Remove class
				case 'class_remove_wrapper' :

					obj_wrapper.removeClass(value);
					debug_action_value = this.language('debug_action_removed');
					break;

				// Add class (Wrapper)
				case 'class_add_field' :

					if(object_row_id !== false) { obj_row.addClass(value); } else { obj.addClass(value); }
					debug_action_value = this.language('debug_action_added');
					break;

				// Remove class
				case 'class_remove_field' :

					if(object_row_id !== false) { obj_row.removeClass(value); } else { obj.removeClass(value); }
					debug_action_value = this.language('debug_action_removed');
					break;
				// Select an option
				case 'value_row_select' :

					obj.find('option:enabled[data-id="' + object_row_id + '"]').prop('selected',false).prop('selected',true);
					debug_action_value = this.language('debug_action_selected');
					break;

				// Deselect an option
				case 'value_row_deselect' :

					obj.find('option:enabled[data-id="' + object_row_id + '"]').prop('selected',true).prop('selected',false);
					debug_action_value = this.language('debug_action_deselected');
					break;

				// Unselect all options (Reset)
				case 'value_row_reset' :

					obj.find('option:enabled').prop('selected',true).prop('selected',false);
					debug_action_value = this.language('debug_action_reset');
					break;

				// Check/uncheck a checkbox or radio
				case 'value_row_check' :
				case 'value_row_uncheck' :

					$(object_selector_row + ':enabled').prop('checked',(action == 'value_row_check'));
					debug_action_value = this.language('debug_action_' + ((action == 'value_row_check') ? 'checked' : 'unchecked'));
					break;

				// Set required on a checkbox or radio
				case 'value_row_required' :
				case 'value_row_not_required' :

					obj_row.prop('required',(action == 'value_row_required'));
					debug_action_value = this.language('debug_action_' + ((action == 'value_row_required') ? 'required' : 'not_required'));
					process_required = true;
					break;

				// Set disabled on a checkbox or radio
				case 'value_row_disabled' :
				case 'value_row_not_disabled' :

					obj_row.attr('disabled',(action == 'value_row_disabled'));
					debug_action_value = this.language('debug_action_' + ((action == 'value_row_disabled') ? 'disabled' : 'enabled'));
					break;

				// Set visible on a checkbox or radio
				case 'value_row_visible' :
				case 'value_row_not_visible' :

					if(action === 'value_row_not_visible') { obj_row.parent().hide(); } else { obj_row.parent().show(); }
					debug_action_value = this.language('debug_action_' + ((action == 'value_row_not_visible') ? 'hide' : 'show'));
					break;

				// Focus on a checkbox or radio
				case 'value_row_focus' :

					obj_row.focus();
					debug_action_value = this.language('debug_action_focussed');
					break;

				// Add class
				case 'value_row_class_add' :

					obj_row.addClass(value);
					debug_action_value = this.language('debug_action_added');
					break;

				// Remove class
				case 'value_row_class_remove' :

					obj_row.removeClass(value);
					debug_action_value = this.language('debug_action_removed');
					break;

				// Set custom validity
				case 'value_row_set_custom_validity' :

					// Custom invalid feedback text
					var invalid_feedback_obj = $('#' + this.form_id_prefix + 'invalid-feedback-' + object_id + '-' + object_row_id);

					// Set invalid feedback
					this.set_invalid_feedback(obj_row, invalid_feedback_obj, value, object_id, object_row_id);

					break;

				// Set visibility
				case 'visibility' :

					switch(object) {

						// Field / section visibility
						case 'section' :
						case 'field' :

							if(value === 'off') { obj_wrapper.hide(); } else { obj_wrapper.show(); }
							debug_action_value = this.language('debug_action_' + ((value == 'off') ? 'hide' : 'show'));
							break;
					}

					break;

				// Disable
				case 'disabled' :

					switch(object) {

						case 'section' :

							// For sections, we need to look for a fieldset
							$('fieldset', obj_wrapper).first().prop('disabled', (value == 'on'));
							break;

						case 'field' :

							obj.prop('disabled', (value == 'on'));

							var class_disabled_array = this.get_field_value_fallback(obj_wrapper.attr('data-type'), false, 'class_disabled', false);

							if(value == 'on') {

								if(class_disabled_array !== false) { obj.addClass(class_disabled_array.join(' ')); }
								obj.css({'pointer-events': 'none'});

							} else {

								if(class_disabled_array !== false) { obj.removeClass(class_disabled_array.join(' ')); }
								obj.attr('style', '');
							}

							break;
					}

					debug_action_value = this.language('debug_action_' + ((value == 'on') ? 'disabled' : 'enabled'));
					break;

				// Required
				case 'required' :

					// Get field data
					var field = this.field_data_cache[object_id];
					switch(field.type) {

						case 'radio' :

							$('[name="' + field_name + '[]"]').prop('required', (value == 'on'));
							break;

						default :

							obj.prop('required', (value == 'on'));
					}

					debug_action_value = this.language('debug_action_' + ((value == 'on') ? 'required' : 'not_required'));
					process_required = true;
					break;

				// Required - Signature
				case 'required_signature' :

					obj.attr('required', (value == 'on'));
					var signature = (typeof(this.signatures_by_name[field_name]) !== 'undefined') ? this.signatures_by_name[field_name] : false;
					if(signature !== false) { signature.required = (value == 'on'); }
					debug_action_value = this.language('debug_action_' + ((value == 'on') ? 'required' : 'not_required'));
					process_required = true;
					break;

				// Read only
				case 'readonly' :

					obj.prop('readonly', (value == 'on'));
					debug_action_value = this.language('debug_action_' + ((value == 'on') ? 'read_only' : 'not_read_only'));
					break;

				// Set custom validity
				case 'set_custom_validity' :

					// Custom invalid feedback text
					var invalid_feedback_obj = $('#' + this.form_id_prefix + 'invalid-feedback-' + object_id);

					// Set invalid feedback
					this.set_invalid_feedback(obj, invalid_feedback_obj, value, object_id);

					break;

				// Click
				case 'click' :

					switch(object) {

						// Tab click
						case 'group' :

							var tab_selector = '[href="#' + this.form_id_prefix + 'tab-content-' + object_id + '"]';
							$(tab_selector).click();
							break;

						// Field click
						case 'field' :

							obj.click();
							break;
					}

					debug_action_value = this.language('debug_action_clicked');
					break;

				// Focus
				case 'focus' :

					obj.focus();
					debug_action_value = this.language('debug_action_focussed');

					break;

				// Action - Run on save
				case 'action_run_on_save' :

					if(this.conditional_actions_run_save.indexOf(object_id) === -1) {
						this.conditional_actions_run_save.push(object_id);
					}
					break;

				// Action - Run on submit
				case 'action_run_on_submit' :

					if(this.conditional_actions_run_submit.indexOf(object_id) === -1) {
						this.conditional_actions_run_submit.push(object_id);
					}
					break;

				// Action - Do not run on save
				case 'action_do_not_run_on_save' :

					var object_id_index = this.conditional_actions_run_save.indexOf(object_id);
					if (object_id_index !== -1) this.conditional_actions_run_save.splice(object_id_index, 1);
					break;

				// Action - Do not run on submit
				case 'action_do_not_run_on_submit' :

					var object_id_index = this.conditional_actions_run_submit.indexOf(object_id);
					if (object_id_index !== -1) this.conditional_actions_run_submit.splice(object_id_index, 1);
					break;
			}

			actions_processed++;

			if($.WS_Form.debug_rendered) {

				// Build action description for debug
				switch(object) {

					case 'group' :

						var object_single = this.group_data_cache[object_id];
						var object_single_type = this.language('debug_action_group');
						break;

					case 'section' :

						var object_single = this.section_data_cache[object_id];
						var object_single_type = this.language('debug_action_section');
						break;

					case 'field' :

						var object_single = this.field_data_cache[object_id];
						var object_single_type = object_single.type;
						break;

					case 'action' :

						var object_single = this.action_data_cache[object_id];
						var object_single_type = this.language('debug_action_action');
						break;

				}

				var conditional_settings = $.WS_Form.settings_form.conditional;
				var conditional_settings_objects = conditional_settings.objects;
				var conditional_settings_actions = conditional_settings_objects[object]['action'];
				var conditional_settings_action = conditional_settings_actions[action];

				var action_description = conditional_settings_action.text.toUpperCase();
				if(typeof conditional_settings_action.values !== 'undefined') {

					if(typeof conditional_settings_action.values === 'object') {

						if(typeof conditional_settings_action.values[value] !== 'undefined') {

							debug_action_value = conditional_settings_action.values[value].text;
						}
					}
				}

				var log_description = '<strong>[' + this.html_encode(object_single.label) + '] ' + action_description + (((debug_action_value !== false) && (debug_action_value != '')) ? " '" + this.html_encode(debug_action_value) + "'" : '') + '</strong> (' + this.language('debug_action_type') + ': ' + object_single_type + ' | ID: ' + object_id + (object_row_id ? ' | ' + this.language('debug_action_row') + ' ID: ' + object_row_id : '') + ')';

				this.log('log_conditional_action_' + type, log_description, 'conditional');
			}
		}

		if(actions_processed == 0) {

			this.log('log_conditional_action_not_found_' + type, '', 'conditional');
		}

		if(process_required) {

			this.form_required();
		}
	}

	$.WS_Form.prototype.conditional_logic_previous = function(accumulator, value, logic_previous) {

		switch(logic_previous) {

			// OR
			case '||' :

				accumulator |= value;
				break;

			// AND
			case '&&' :

				accumulator &= value;
				break;
		}

		return accumulator;
	}

	// Check integrity of a condition
	$.WS_Form.prototype.conditional_condition_check = function(condition) {

		// Check condition structure
		if(typeof condition.id === 'undefined') { return false; }
		if(typeof condition.object === 'undefined') { return false; }
		if(typeof condition.object_id === 'undefined') { return false; }
		if(typeof condition.object_row_id === 'undefined') { return false; }
		if(typeof condition.logic === 'undefined') { return false; }
		if(typeof condition.value === 'undefined') { return false; }
		if(typeof condition.case_sensitive === 'undefined') { return false; }
		if(typeof condition.logic_previous === 'undefined') { return false; }

		// Check condition variables
		if(condition.id == '') { return false; }
		if(condition.id == 0) { return false; }
		if(condition.object == '') { return false; }
		if(condition.object_id == '') { return false; }
		if(condition.logic == '') { return false; }

		return true;
	}

	// Check integrity of an action
	$.WS_Form.prototype.conditional_action_check = function(action) {

		// Check action structure
		if(typeof action.object === 'undefined') { return false; }
		if(typeof action.object_id === 'undefined') { return false; }
		if(typeof action.action === 'undefined') { return false; }

		// Check action variables
		if(action.object == '') { return false; }
		if(action.object_id == '') { return false; }
		if(action.action == '') { return false; }

		return true;
	}

	// Group - Tabs - Init
	$.WS_Form.prototype.form_tabs = function() {

		if(this.form.groups.length <= 1) { return false; }

		var ws_this = this;

		// Get tab index cookie if settings require it
		var index = (this.get_object_meta_value(this.form, 'cookie_tab_index')) ? this.cookie_get('tab_index', 0) : 0;

		// If we are using the WS Form framework, then we need to run our own tabs script
		if($.WS_Form.settings_plugin.framework == ws_form_settings.framework_admin) {

			var tabs_obj = $('.wsf-group-tabs', this.form_obj);

			// Destroy tabs (Ensures subsequent calls work)
			if(tabs_obj.hasClass('wsf-tabs')) { this.tabs_destroy(); }

			// Init tabs
			this.tabs(tabs_obj, { active: index });

		} else {

			// Set active tab
			this.group_index_set(index);
		}

		// Set up on click event for each tab
		$('[href^="#' + this.form_id_prefix + 'tab-content-"]').click(function () {

			var tab_index = $(this).parent().index();

			// Save current tab index to cookie
			if(ws_this.get_object_meta_value(ws_this.form, 'cookie_tab_index')) {

				ws_this.cookie_set('tab_index', tab_index);
			}

		});
	}

	// Form - Submit
	$.WS_Form.prototype.form_post = function(submit) {

		// Determine if this is a submit or a save
		if(typeof submit == 'undefined') { var submit = false; }

		// Trigger rendered event
		this.trigger(submit ? 'submit' : 'save');

		var ws_this = this;

		// Lock form
		this.form_post_lock();

		// Build form data
		var form_data = new FormData(this.form_obj[0]);
		form_data.append('id', this.form_obj.attr('data-id'));
		form_data.append('hash', this.hash);
		form_data.append('duration', Math.round((new Date().getTime() - this.date_start) / 1000));

		if((typeof(ws_form_settings.post_id) !== 'undefined') && (ws_form_settings.post_id > 0)) {

			form_data.append('post_id', ws_form_settings.post_id);
		}

		if(submit) { form_data.append('submit', 'yes'); }

		// Safari bug fix (Empty file fields cause Safari to bug out)
		$('input[type=file]', this.form_obj).each(function() {

			var files = $(this).prop('files');

			if(files != undefined && files.length <= 0) { form_data.delete($(this).attr('name')); }
		});

 		this.api_call('submit', 'POST', form_data, function(response) {

 			// Success
			ws_this.form_post_unlock();

			// Clear hash
			ws_this.form_hash_clear();

			// Check for validation errors
			var error_validation = (typeof(response.error_validation) !== 'undefined') && response.error_validation;

			// Check for errors
			var errors = (typeof(response.data.errors) !== 'undefined') && response.data.errors.length;

 			// Check for form reload on submit
 			if(submit && !error_validation && !errors && ws_this.get_object_meta_value(ws_this.form, 'submit_reload', true)) {

 				// Reload
 				ws_this.form_reload();
	 		}

 		}, function() {

 			// Error
			ws_this.form_post_unlock();

 		});
	}

	// Form lock
	$.WS_Form.prototype.form_post_lock = function() {

		// Trigger rendered event
		this.trigger('lock');

		if(this.get_object_meta_value(this.form, 'submit_lock', false)) {

			// Disable post buttons
			this.form_post_locked = true;

		}
	}

	// Form unlock
	$.WS_Form.prototype.form_post_unlock = function() {

		var ws_this = this;

		if(this.get_object_meta_value(this.form, 'submit_lock', false)) {

			// Enable post buttons
			setTimeout(function() {

				ws_this.form_post_locked = false;


				// Trigger rendered event
				ws_this.trigger('unlock');

			}, 1000);
		}
	}

	// API Call
	$.WS_Form.prototype.api_call = function(ajax_path, method, params, success_callback, error_callback) {


		// Defaults
		if(typeof method === 'undefined') { var method = 'POST'; }
		if(typeof params === 'undefined') { var params = false; }

		// Make AJAX request
		var url = ws_form_settings.url + ajax_path;

		var ajax_http_method_override = ($.WS_Form.settings_plugin.ajax_http_method_override === true);

		var ws_this = this;
		$.ajax({

			method: ajax_http_method_override ? 'POST' : method,
			url: url,
			beforeSend: function(xhr) {

				xhr.setRequestHeader('X-WP-Nonce', ws_form_settings.nonce);

				if(ajax_http_method_override) {

					xhr.setRequestHeader('X-HTTP-Method-Override', method);
				}
			},
			data: params,
			contentType: false,
			processData: false,
  			statusCode: {

				// Success
				200: function(response) {

					// Handle hash response
					var hash_ok = ws_this.api_call_hash(response);

					// Call success function
					if(typeof success_callback !== 'undefined') { success_callback(response); } else { ws_this.loader_off(); }

					if(typeof(response.data) !== 'undefined') {

						// Check for nonce
						if(typeof(response.data.nonce) !== 'undefined') { ws_form_settings.nonce = response.data.nonce; }

						// Check for action_js (These are returned from the action system to tell the browser to do something)
						if(typeof(response.data.js) === 'object') { ws_this.action_js_init(response.data.js); }
					}
				},

				// Bad request (Error from API)
				400: function(response) {

					var data = response.responseJSON;


					if(typeof error_callback !== 'undefined') { error_callback(data); } else { ws_this.loader_off(); }
				},

				// Forbidden (Error from API)
				403: function(response) {

					var data = response.responseJSON;


					if(typeof error_callback !== 'undefined') { error_callback(data); } else { ws_this.loader_off(); }
				},

				// Not found (Error from API)
				404: function(response) {

					var data = response.responseJSON;


					if(typeof error_callback !== 'undefined') { error_callback(data); } else { ws_this.loader_off(); }
				},

				// Server error
				500: function(response) {

					var data = response.responseJSON;


					if(typeof error_callback !== 'undefined') { error_callback(data); } else { ws_this.loader_off(); }
				}
			}
		});

		return this;
	};

	// JS Actions - Init
	$.WS_Form.prototype.action_js_init = function(action_js) {

		// Trigger rendered event
		this.trigger('actions-start');

		this.action_js = action_js;

		this.action_js_process_next();
	};

	$.WS_Form.prototype.action_js_process_next = function() {

		if(this.action_js.length == 0) {

		 	// Trigger rendered event
			this.trigger('actions-finish');

			return false;
		}

		var js_action = this.action_js.shift();

		var action = this.js_action_get_parameter(js_action, 'action');
		var duration = this.js_action_get_parameter(js_action, 'duration');

		switch(action) {

			// Redirect
			case 'redirect' :

				var url = this.js_action_get_parameter(js_action, 'url');
				if(url !== false) { location.href = js_action['url']; }

				// Actions end at this point because of the redirect

				break;

			// Message
			case 'message' :

				var message = this.js_action_get_parameter(js_action, 'message');
				var type = this.js_action_get_parameter(js_action, 'type');
				var method = this.js_action_get_parameter(js_action, 'method');
				var form_hide = this.js_action_get_parameter(js_action, 'form_hide');
				var clear = this.js_action_get_parameter(js_action, 'clear');
				var scroll_top = this.js_action_get_parameter(js_action, 'scroll_top');
				var form_show = this.js_action_get_parameter(js_action, 'form_show');
				var message_hide = this.js_action_get_parameter(js_action, 'message_hide');

				this.action_message(message, type, method, duration, form_hide, clear, scroll_top, form_show, message_hide);

				break;

			// Unknown
			default :

				this.action_js_process_next();
		}
	}

	// JS Actions - Get js_action config parameter from AJAX return
	$.WS_Form.prototype.js_action_get_parameter = function(js_action_parameters, meta_key) {

		return typeof(js_action_parameters[meta_key]) !== 'undefined' ? js_action_parameters[meta_key] : false;
	}

	// JS Actions - Get framework config value
	$.WS_Form.prototype.get_framework_config_value = function(object, meta_key) {

		if(typeof(this.framework[object]) === 'undefined') {
			return false;
		}
		if(typeof(this.framework[object]['public']) === 'undefined') {
			return false;
		}
		if(typeof(this.framework[object]['public'][meta_key]) === 'undefined') { return false; }

		return this.framework[object]['public'][meta_key];
	}


	// JS Action - Message
	$.WS_Form.prototype.action_message = function(message, type, method, duration, form_hide, clear, scroll_top, form_show, message_hide) {

		if(typeof(type) === 'undefined') { var type = 'information'; }
		if(typeof(method) === 'undefined') { var method = 'before'; }
		if(typeof(duration) === 'undefined') { var duration = 0; }
		if(typeof(form_hide) === 'undefined') { var form_hide = false; }
		if(typeof(clear) === 'undefined') { var clear = true; }
		if(typeof(scroll_top) === 'undefined') { var scroll_top = true; }
		if(typeof(form_show) === 'undefined') { var form_show = false; }
		if(typeof(message_hide) === 'undefined') { var message_hide = false; }

		// Parse duration
		duration = parseInt(duration);
		if(duration < 0) { duration = 0; }

		// Get config
		var mask_wrapper = this.get_framework_config_value('message', 'mask_wrapper');
		var types = this.get_framework_config_value('message', 'types');

		var type = (typeof(types[type]) !== 'undefined') ? types[type] : false;
		var mask_wrapper_class = (typeof(type['mask_wrapper_class']) !== 'undefined') ? type['mask_wrapper_class'] : false;

		// Clear other messages
		if(clear) {

			$('[data-wsf-message]').remove();
		}

		// Scroll top
		if(scroll_top) {

			window.scrollTo(0,0);
		}

		var mask_wrapper_values = {

			'message': 				message,
			'mask_wrapper_class': 	mask_wrapper_class 
		};

		var message_div = $('<div/>', { html: this.mask_parse(mask_wrapper, mask_wrapper_values) });
		message_div.attr('data-wsf-message', '');

		// Hide form?
		if(form_hide) { this.form_obj.hide(); }

		// Render message
		switch(method) {

			// Before
			case 'before' :

				message_div.insertBefore(this.form_obj);
				break;

			// After
			case 'after' :

				message_div.insertAfter(this.form_obj);
				break;
		}

		// Process next action
		var ws_this = this;

		setTimeout(function() {

			// Should this message be removed?
			if(message_hide) { message_div.remove(); }

			// Should the form be shown?
			if(form_show) { ws_this.form_obj.show(); }

			// Process next js_action
			ws_this.action_js_process_next();

		}, parseInt(duration));
	}

	// Text input and textarea character and word count
	$.WS_Form.prototype.form_character_word_count = function(obj) {

		var ws_this = this;
		if(typeof(obj) === 'undefined') { var obj = this.form_obj; }

		// Run through each input that accepts text
		for(var field_id in this.field_data_cache) {

			var field = this.field_data_cache[field_id];

			if(

				// Check if field has these meta keys
				this.has_object_meta_key(field, 'min_length') ||
				this.has_object_meta_key(field, 'max_length') ||
				this.has_object_meta_key(field, 'min_length_words') ||
				this.has_object_meta_key(field, 'max_length_words')
			) {

				// Process count functionality on field
				var field_obj = $('#' + this.form_id_prefix + 'field-' + field_id, obj);
				if(!field_obj.length) { continue; }

				if(this.form_character_word_count_process(field_obj)) {

					field_obj.on('keyup change focus blur paste', function() { ws_this.form_character_word_count_process($(this)); });
				}
			}
		}
	}

	// Text input and textarea character and word count - Process
	$.WS_Form.prototype.form_character_word_count_process = function(obj) {

		// Get help text
		var field_wrapper_obj = obj.closest('[data-type]');
		var field_id = field_wrapper_obj.attr('data-id');
		var field = this.field_data_cache[field_id];
		var help = this.get_object_meta_value(field, 'help', '', false, true);

		// If #character_ and #word_ not present, don't bother processing
		if(
			(help.indexOf('#character_') === -1) &&
			(help.indexOf('#word_') === -1)
		) {
			return false;
		}

		// Get minimum and maximum character count
		var min_length = this.get_object_meta_value(field, 'min_length', '');
		min_length = (parseInt(min_length) > 0) ? parseInt(min_length) : false;

		var max_length = this.get_object_meta_value(field, 'max_length', '');
		max_length = (parseInt(max_length) > 0) ? parseInt(max_length) : false;

		// Get minimum and maximum word length
		var min_length_words = this.get_object_meta_value(field, 'min_length_words', '');
		min_length_words = (parseInt(min_length_words) > 0) ? parseInt(min_length_words) : false;

		var max_length_words = this.get_object_meta_value(field, 'max_length_words', '');
		max_length_words = (parseInt(max_length_words) > 0) ? parseInt(max_length_words) : false;

		// Calculate sizes
		var val = obj.val();

		var character_count = val.length;
		var character_remaining = (max_length !== false) ? max_length - character_count : false;
		if(character_remaining < 0) { character_remaining = 0; }

		var word_count = this.get_word_count(val);
		var word_remaining = (max_length_words !== false) ? max_length_words - word_count : false;
		if(word_remaining < 0) { word_remaining = 0; }

		// Check minimum and maximums counts
		var count_invalid = false;
		var count_invalid_message_array = [];

		if((min_length !== false) && (character_count < min_length)) {

			count_invalid_message_array.push(this.language('error_min_length', min_length));
			count_invalid = true;
		}
		if((max_length !== false) && (character_count > max_length)) {

			count_invalid_message_array.push(this.language('error_max_length', max_length));
			count_invalid = true;
		}
		if((min_length_words !== false) && (word_count < min_length_words)) {

			count_invalid_message_array.push(this.language('error_min_length_words', min_length_words));
			count_invalid = true;
		}
		if((max_length_words !== false) && (word_count > max_length_words)) {

			count_invalid_message_array.push(this.language('error_max_length_words', max_length_words));
			count_invalid = true;
		}

		// Get object ID
		var object_id = obj.closest('[data-id]').attr('data-id');

		// Custom invalid feedback text
		var invalid_feedback_obj = $('#' + this.form_id_prefix + 'invalid-feedback-' + object_id);

		// Check if count_invalid
		if(count_invalid) {

			// Set invalid feedback
			this.set_invalid_feedback(obj, invalid_feedback_obj, count_invalid_message_array.join(' / '), object_id);

		} else {

			// Reset invalid feedback
			this.set_invalid_feedback(obj, invalid_feedback_obj, '', object_id);
		}

		// Get language
		var character_singular = this.language('character_singular');
		var character_plural = this.language('character_plural');
		var word_singular = this.language('word_singular');
		var word_plural = this.language('word_plural');

		// Set mask values
		var mask_values_help = {

			// Characters
			'character_count': 				character_count,
			'character_count_label': 		(character_count == 1 ? character_singular : character_plural),
			'character_remaining': 			(character_remaining !== false) ? character_remaining : '',
			'character_remaining_label': 	(character_remaining == 1 ? character_singular : character_plural),
			'character_min': 				(min_length !== false) ? min_length : '',
			'character_min_label': 			(min_length !== false) ? (min_length == 1 ? character_singular : character_plural) : '',
			'character_max': 				(max_length !== false) ? max_length : '',
			'character_max_label': 			(max_length !== false) ? (max_length == 1 ? character_singular : character_plural) : '',

			// Words
			'word_count': 			word_count,
			'word_count_label': 	(word_count == 1 ? word_singular : word_plural),
			'word_remaining': 		(word_remaining !== false) ? word_remaining : '',
			'word_remaining_label': (word_remaining == 1 ? word_singular : word_plural),
			'word_min': 			(min_length_words !== false) ? min_length_words : '',
			'word_min_label': 		(min_length_words !== false) ? (min_length_words == 1 ? word_singular : word_plural) : '',
			'word_max': 			(max_length_words !== false) ? max_length_words : '',
			'word_max_label': 		(max_length_words !== false) ? (max_length_words == 1 ? word_singular : word_plural) : ''
		};

		// Parse help mask
		var help_parsed = this.mask_parse(help, mask_values_help);

		// Update help HTML
		var help_id = this.form_id_prefix + 'help-' + field.id;
		$('#' + help_id).html(help_parsed);

		return true;
	}

	// Get word count of a string
	$.WS_Form.prototype.get_word_count = function(input_string) {

		// Trim input string
		input_string = input_string.trim();

		// If string is empty, return 0
		if(input_string.length == 0) { return 0; }

		// Return word count
		return input_string.trim().replace(/\s+/gi, ' ').split(' ').length;
	}

	// API Call
	$.WS_Form.prototype.api_call_hash = function(response) {

		var hash_ok = true;
		if(typeof response.hash === 'undefined') { hash_ok = false; }
		if(hash_ok && (response.hash.length != 32)) { hash_ok = false; }
		if(hash_ok) {

			// Set hash
			this.hash_set(response.hash)
		}

		return hash_ok;
	}

	// Hash - Set
	$.WS_Form.prototype.hash_set = function(hash, cookie_set) {

		if(typeof cookie_set === 'undefined') { cookie_set = false; }

		if(hash != this.hash) {

			// Set hash
			this.hash = hash;

			// Set hash cookie
			cookie_set = true;


			// Session session on form tag
			this.form_obj.attr('hash', this.hash);
		}

		if(cookie_set) {

			this.cookie_set('hash', this.hash);
		}
	}

	// Generate password
	$.WS_Form.prototype.generate_password = function(length) {

		var password = '';
		var characters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*()_+~`|}{[]\\:;?><,./-=';
		
		for(var i = 0; i < length; ++i) { password += characters.charAt(Math.floor(Math.random() * characters.length)); }

		return password;
	}

	// On load
	$(function() {

		window.wsf_form_instances = [];

		// Render each form
		$('.wsf-form').each(function() {

			var id = $(this).attr('id');
			var instance_id = $(this).attr('data-instance-id');
			var form_id = $(this).attr('data-id');
			var form_index = $(this).attr('data-index');

			var ws_form = new $.WS_Form();
			window.wsf_form_instances[instance_id] = ws_form;

			ws_form.render({

				'obj' : 		'#' + id,
				'form_id':		form_id
			});
		});
	});

})(jQuery);


